#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2020 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------


__all__ = ["register_with_opt_einsum"]


def _dispatch_diagonal(subscript, outstring, operand):
    # Do the diagonal call one index at a time.
    char = None
    for c in subscript:
        if subscript.count(c) > 1:
            char = c
            break
    assert char is not None
    indices = [i for (i, c) in enumerate(subscript) if c == char]

    # Diagonal pushes the diagonalised index back
    newoperand = operand.diagonal(*indices)
    newsubscript = "".join(c for c in subscript if c != char) + char
    if len(newsubscript) > len(outstring):
        # Recurse if more letters for which diagonals are to be extracted
        return _dispatch_diagonal(newsubscript, outstring, newoperand)
    elif newsubscript == outstring:
        return newoperand
    else:
        # Transpose newsubscript -> outstring
        permutation = tuple(map(newsubscript.index, outstring))
        return newoperand.transpose(permutation)


def _fallback_einsum(einsum_str, *operands, **kwargs):
    # A fallback implementation of einsum in adcc,
    # which deals with a few cases opt_einsum cannot deal with
    from .functions import einsum
    from opt_einsum.parser import gen_unused_symbols

    operands = list(operands)
    subscripts = einsum_str.split("->")[0].split(",")
    outstr = einsum_str.split("->")[1]

    # If there are any diagonal extractions, which can be done in the operands,
    # do them first.
    for i in range(len(subscripts)):
        sub = subscripts[i]
        cdiagonal = set(c for c in sub if sub.count(c) > 1 and c in outstr)
        ctrace = set(c for c in sub if sub.count(c) > 1 and c not in outstr)
        if ctrace:
            raise NotImplementedError("Partial traces (e.g. contractions "
                                      "'iaib->ab') are not yet supported "
                                      "in adcc.einsum.")
        if cdiagonal:
            # Do any possible diagonal extraction first
            outstring = "".join(c for c in sub if c not in cdiagonal)
            outstring += "".join(cdiagonal)
            operands[i] = _dispatch_diagonal(subscripts[i], outstring,
                                             operands[i])
            subscripts[i] = outstring

    if len(subscripts) == 1:
        # At this point all which is left should be a permutation.
        assert all(c in outstr for c in subscripts[0])
        permutation = tuple(map(subscripts[0].index, outstr))
        return operands[0].transpose(permutation)
    elif len(subscripts) == 2:
        # Should the diagonal of a contraction be extracted, e.g. il,laib->aib
        diagonal_chars = [a for a in subscripts[0]
                          if a in subscripts[1] and a in outstr]

        if not diagonal_chars:
            # Try another round of einsum
            return einsum(",".join(subscripts) + "->" + outstr, *operands)

        # Replace one of the duplicate characters in the input
        # and prepend it to output
        replacers = list(gen_unused_symbols(outstr + "".join(subscripts),
                                            len(diagonal_chars)))
        newoutstr = "".join(replacers) + outstr
        for (old, new) in zip(diagonal_chars, replacers):
            subscripts[0] = subscripts[0].replace(old, new)

        # Check we are not creating an infinite loop:
        assert ",".join(subscripts) + "->" + newoutstr != einsum_str

        # Run einsum doing the partial contraction
        # (well actually we should directly do tensordot)
        res = einsum(",".join(subscripts) + "->" + newoutstr, *operands)

        # Run _dispatch_diagonal with the result to form the requested diagonal
        return _dispatch_diagonal("".join(diagonal_chars) + outstr, outstr, res)
    else:
        raise NotImplementedError("Fallback einsum not implemented for more than "
                                  "two operators")


def register_with_opt_einsum():
    import libadcc

    from opt_einsum.backends.dispatch import (EVAL_CONSTS_BACKENDS,
                                              _cached_funcs, _has_einsum)

    def libadcc_evaluate_constants(const_arrays, expr):
        # Compute the partial expression tree of the inputs
        new_ops, new_contraction_list = expr(*const_arrays, backend='libadcc',
                                             evaluate_constants=True)

        # Evaluate as much as possible and return
        new_ops = [None if x is None else libadcc.evaluate(x) for x in new_ops]
        return new_ops, new_contraction_list
    EVAL_CONSTS_BACKENDS["libadcc"] = libadcc_evaluate_constants
    _has_einsum["libadcc"] = False
    _cached_funcs[('einsum', 'libadcc')] = _fallback_einsum



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2019 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
import copy

from .Tensor import _tensor_select_below_absmax
from .FormatIndex import FormatIndexAdcc, FormatIndexBase

from libadcc import Tensor


class FormatDominantElements:
    def __init__(self, mospaces, tolerance=0.01, index_format=FormatIndexAdcc):
        self.mospaces = mospaces
        self.tolerance = tolerance
        self.value_format = "{:+8.3g}"  # Formatting used for the values

        if isinstance(index_format, type):
            self.index_format = index_format(self.mospaces)
        elif isinstance(index_format, FormatIndexBase):
            self.index_format = copy.copy(index_format)
        else:
            raise TypeError("index_format needs to be of type FormatIndexBase")

    def optimise_formatting(self, *spaces_tensor_pairs: tuple[tuple[str], Tensor]):
        """
        Optimise the formatting parameters of this class and the index_format
        class in order to be able to nicely produce equivalently formatted tensor
        format strings for all the passed spaces-tensor pairs.

        This function can be called multiple times.
        """
        assert all(len(pair) == 2 for pair in spaces_tensor_pairs)

        for spaces, tensor in spaces_tensor_pairs:
            if not isinstance(spaces, (tuple, list)) or \
               not isinstance(tensor, Tensor):
                raise TypeError("spaces_tensor_pairs should be a list of "
                                "(spaces tuples, Tensor) tuples")
            for indices, _ in _tensor_select_below_absmax(tensor, self.tolerance):
                assert len(indices) == len(spaces)
                self.index_format.optimise_formatting(
                    *((sp, idx) for sp, idx in zip(spaces, indices))
                )

    def format_as_list(self, spaces: tuple[str], tensor: Tensor) -> list[str]:
        """
        Raw-format the dominant tensor elements as a list of tuples with one
        tuple for each element. Each tuple has three entries, the formatted
        indices, the formatted spins and the value
        """
        ret = []
        for indices, value in _tensor_select_below_absmax(tensor, self.tolerance):
            assert len(indices) == len(spaces)
            formatted = tuple(self.index_format.format(sp, idx,
                                                       concat_spin=False)
                              for sp, idx in zip(spaces, indices))
            ret.append(tuple(zip(*formatted)) + (value, ))
        return ret

    def format(self, spaces: tuple[str], tensor: Tensor) -> str:
        """
        Return a multiline string representing the dominant tensor elements.
        The tensor index is formatted according to the index format passed
        upon class construction.
        """
        ret = []
        for indices, spins, value in self.format_as_list(spaces, tensor):
            ret.append(" ".join(indices) + "  " + "".join(spins)
                       + "   " + self.value_format.format(value))
        return "\n".join(ret)



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2018 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
import numpy as np

from libadcc import HartreeFockProvider

from .EriBuilder import EriBuilder
from ..exceptions import InvalidReference
from ..ElectronicStates import EnergyCorrection
from ..OneParticleOperator import OneParticleOperator

from pyscf import ao2mo, gto, scf
from pyscf.solvent import ddcosmo


class PyScfOperatorIntegralProvider:
    available: tuple[str, ...] = (
        "overlap", "electric_dipole", "electric_dipole_velocity", "magnetic_dipole",
        "electric_quadrupole", "electric_quadrupole_traceless",
        "electric_quadrupole_velocity", "diamagnetic_magnetizability",
        "pe_induction_elec", "pcm_potential_elec"
    )

    def __init__(self, scfres):
        self.scfres = scfres
        self.backend = "pyscf"

    @property
    def overlap(self) -> np.ndarray:
        return self.scfres.mol.intor_symmetric('int1e_ovlp')

    @property
    def electric_dipole(self) -> tuple[np.ndarray, ...]:
        """-sum_i r_i"""
        return tuple(
            -1.0 * self.scfres.mol.intor_symmetric('int1e_r', comp=3)
        )

    def magnetic_dipole(self, gauge_origin="origin") -> tuple[np.ndarray, ...]:
        """
        The imaginary part of the integral is returned.
        -0.5 * sum_i r_i x p_i
        """
        gauge_origin = _transform_gauge_origin_to_xyz(self.scfres, gauge_origin)
        with self.scfres.mol.with_common_orig(gauge_origin):
            return tuple(
                -0.5 * self.scfres.mol.intor('int1e_cg_irxp', comp=3, hermi=2)
            )

    @property
    def electric_dipole_velocity(self) -> tuple[np.ndarray, ...]:
        """
        The imaginary part of the integral is returned.
        -sum_i p_i
        """
        with self.scfres.mol.with_common_orig((0.0, 0.0, 0.0)):
            return tuple(
                self.scfres.mol.intor('int1e_ipovlp', comp=3, hermi=2)
            )

    def electric_quadrupole(self, gauge_origin="origin") -> tuple[np.ndarray, ...]:
        """-sum_i r_{i, alpha} r_{i, beta}"""
        gauge_origin = _transform_gauge_origin_to_xyz(self.scfres, gauge_origin)
        with self.scfres.mol.with_common_orig(gauge_origin):
            return tuple(
                -1.0 * self.scfres.mol.intor_symmetric('int1e_rr', comp=9)
            )

    def electric_quadrupole_traceless(self, gauge_origin="origin"
                                      ) -> tuple[np.ndarray, ...]:
        """
        -0.5 * sum_i (3 * r_{i, alpha} r_{i, beta}
        - delta_{alpha, beta} r_{i}^2)
        """
        gauge_origin = _transform_gauge_origin_to_xyz(self.scfres, gauge_origin)
        with self.scfres.mol.with_common_orig(gauge_origin):
            r_r = self.scfres.mol.intor_symmetric('int1e_rr', comp=9)
            r_r = np.reshape(r_r, (3, 3, r_r.shape[1], r_r.shape[1]))
            r_quadr = self.scfres.mol.intor_symmetric('int1e_r2', comp=1)
            r_quadr_matrix = np.zeros_like(r_r)
            for i in range(3):
                r_quadr_matrix[i][i] = r_quadr
            term = 0.5 * (3 * r_r - r_quadr_matrix)
            return tuple(
                -1.0 * np.reshape(term, (9, r_quadr.shape[0], r_quadr.shape[0]))
            )

    def electric_quadrupole_velocity(self, gauge_origin="origin"
                                     ) -> tuple[np.ndarray, ...]:
        """
        The imaginary part of the integral is returned.
        -sum_i (r_{i, beta} p_{i, alpha} - i delta_{alpha, beta}
        + r_{i, alpha} p_{i, beta})
        """
        gauge_origin = _transform_gauge_origin_to_xyz(self.scfres, gauge_origin)
        with self.scfres.mol.with_common_orig(gauge_origin):
            r_p = self.scfres.mol.intor('int1e_irp', comp=9, hermi=0)
            r_p = np.reshape(r_p, (3, 3, r_p.shape[1], r_p.shape[1]))
            ovlp = self.scfres.mol.intor_symmetric('int1e_ovlp', comp=1)
            ovlp_matrix = np.zeros_like(r_p)
            for i in range(3):
                ovlp_matrix[i][i] = ovlp
            term = r_p + r_p.transpose(1, 0, 2, 3) - ovlp_matrix
            return tuple(
                -1.0 * np.reshape(term, (9, ovlp.shape[0], ovlp.shape[0]))
            )

    def diamagnetic_magnetizability(self, gauge_origin="origin"
                                    ) -> tuple[np.ndarray, ...]:
        """
        0.25 * sum_i (r_{i, alpha} r_{i, beta}
        - delta_{alpha, beta} r_{i}^2)
        """
        gauge_origin = _transform_gauge_origin_to_xyz(self.scfres, gauge_origin)
        with self.scfres.mol.with_common_orig(gauge_origin):
            r_r = self.scfres.mol.intor_symmetric('int1e_rr', comp=9)
            r_r = np.reshape(r_r, (3, 3, r_r.shape[1], r_r.shape[1]))
            r_quadr = self.scfres.mol.intor_symmetric('int1e_r2', comp=1)
            r_quadr_matrix = np.zeros_like(r_r)
            for i in range(3):
                r_quadr_matrix[i][i] = r_quadr
            term = 0.25 * (r_quadr_matrix - r_r)
            return tuple(
                np.reshape(term, (9, r_quadr.shape[0], r_quadr.shape[0]))
            )

    def pe_induction_elec(self, dm: OneParticleOperator) -> np.ndarray:
        try:
            self.scfres.with_solvent.cppe_state
        except AttributeError:
            raise RuntimeError("Can not compute the PE electronic induction "
                               "operator using the given pyscf object.")

        return self.scfres.with_solvent._exec_cppe(
            dm.to_ndarray(), elec_only=True
        )[1]

    def pcm_potential_elec(self, dm: OneParticleOperator) -> np.ndarray:
        if not hasattr(self.scfres, "with_solvent") or \
                not isinstance(self.scfres.with_solvent, ddcosmo.DDCOSMO):
            raise RuntimeError("Can not compute the electronic PCM potential "
                               "operator using the given pyscf object.")

        # Since eps (dielectric constant) is the only solvent parameter
        # in pyscf and there is no solvent data available in the
        # program, the user needs to adjust scfres.with_solvent.eps
        # manually to the optical dielectric constant (if non
        # equilibrium solvation is desired).
        return self.scfres.with_solvent._B_dot_x(dm.to_ndarray())


# TODO: refactor ERI builder to be more general
# IntegralBuilder would be good
class PyScfEriBuilder(EriBuilder):
    def __init__(self, scfres, n_orbs, n_orbs_alpha, n_alpha, n_beta, restricted):
        self.scfres = scfres
        if restricted:
            self.mo_coeff = (self.scfres.mo_coeff, self.scfres.mo_coeff)
        else:
            self.mo_coeff = self.scfres.mo_coeff
        super().__init__(n_orbs, n_orbs_alpha, n_alpha, n_beta, restricted)

    @property
    def coefficients(self):
        return {
            "Oa": self.mo_coeff[0][:, :self.n_alpha],
            "Ob": self.mo_coeff[1][:, :self.n_beta],
            "Va": self.mo_coeff[0][:, self.n_alpha:],
            "Vb": self.mo_coeff[1][:, self.n_beta:],
        }

    def compute_mo_eri(self, blocks, spins):
        coeffs = tuple(self.coefficients[blocks[i] + spins[i]] for i in range(4))
        # TODO Pyscf uses HDF5 internal to do the AO2MO here we read it all
        #      into memory. This wastes memory and could be avoided if temporary
        #      files were used instead. These could be deleted on the call
        #      to `flush_cache` automatically.
        sizes = [i.shape[1] for i in coeffs]
        return ao2mo.general(self.scfres.mol, coeffs,
                             compact=False).reshape(sizes[0], sizes[1],
                                                    sizes[2], sizes[3])


class PyScfHFProvider(HartreeFockProvider):
    """
        This implementation is only valid
        if no orbital reordering is required.
    """
    def __init__(self, scfres):
        # Do not forget the next line,
        # otherwise weird errors result
        super().__init__()
        self.scfres = scfres
        n_alpha, n_beta = scfres.mol.nelec
        self.eri_builder = PyScfEriBuilder(self.scfres, self.n_orbs,
                                           self.n_orbs_alpha, n_alpha,
                                           n_beta, self.restricted)
        self.operator_integral_provider = PyScfOperatorIntegralProvider(
            self.scfres
        )
        if not self.restricted:
            assert self.scfres.mo_coeff[0].shape[1] == \
                self.scfres.mo_coeff[1].shape[1]

        self.environment = None
        self.environment_implementation = None
        if hasattr(self.scfres, "with_solvent"):
            if hasattr(self.scfres.with_solvent, "cppe_state"):
                self.environment = "pe"
                self.environment_implementation = "cppe"
            elif isinstance(self.scfres.with_solvent, ddcosmo.DDCOSMO):
                self.environment = "pcm"
                self.environment_implementation = "ddcosmo"

    def pe_energy(self, dm, elec_only=True):
        pe_state = self.scfres.with_solvent
        e_pe, _ = pe_state.kernel(dm.to_ndarray(), elec_only=elec_only)
        return e_pe

    def pcm_energy(self, dm):
        # Since eps (dielectric constant) is the only solvent parameter
        # in pyscf and there is no solvent data available in the
        # program, the user needs to adjust scfres.with_solvent.eps
        # manually to the optical dielectric constant (if non
        # equilibrium solvation is desired).
        V_pcm = self.scfres.with_solvent._B_dot_x(dm.to_ndarray())
        return np.einsum("uv,uv->", dm.to_ndarray(), V_pcm)

    @property
    def excitation_energy_corrections(self):
        ret = []
        if self.environment == "pe":
            ptlr = EnergyCorrection(
                "pe_ptlr_correction",
                lambda view: 2.0 * self.pe_energy(view.transition_dm_ao,
                                                  elec_only=True)
            )
            ptss = EnergyCorrection(
                "pe_ptss_correction",
                lambda view: self.pe_energy(view.state_diffdm_ao,
                                            elec_only=True)
            )
            ret.extend([ptlr, ptss])
        elif self.environment == "pcm":
            ptlr = EnergyCorrection(
                "pcm_ptlr_correction",
                lambda view: self.pcm_energy(view.transition_dm_ao)
            )
            ret.extend([ptlr])
        return {ec.name: ec for ec in ret}

    def get_backend(self):
        return "pyscf"

    def get_conv_tol(self):
        if self.scfres.conv_tol_grad is None:
            conv_tol = self.scfres.conv_tol
        else:
            conv_tol = max(self.scfres.conv_tol, self.scfres.conv_tol_grad**2)
        return conv_tol

    def get_restricted(self):
        if isinstance(self.scfres.mo_occ, list):
            restricted = len(self.scfres.mo_occ) < 2
        elif isinstance(self.scfres.mo_occ, np.ndarray):
            restricted = self.scfres.mo_occ.ndim < 2
        else:
            raise InvalidReference("Unusual pyscf SCF class encountered. Could "
                                   "not determine restricted / unrestricted.")
        return restricted

    def get_energy_scf(self):
        return float(self.scfres.e_tot)

    def get_spin_multiplicity(self):
        # Note: In the pyscf world spin is 2S, so the multiplicity
        #       is spin + 1
        return int(self.scfres.mol.spin) + 1

    def get_n_orbs_alpha(self):
        if self.restricted:
            return self.scfres.mo_coeff.shape[1]
        else:
            return self.scfres.mo_coeff[0].shape[1]

    def get_n_bas(self):
        return int(self.scfres.mol.nao_nr())

    def get_nuclear_multipole(self, order, gauge_origin=(0, 0, 0)):
        charges = self.scfres.mol.atom_charges()
        if order == 0:
            # The function interface needs to be a np.array on return
            return np.array([np.sum(charges)])
        elif order == 1:
            coords = self.scfres.mol.atom_coords()
            return np.einsum("i,ix->x", charges, coords)
        elif order == 2:
            coords = self.scfres.mol.atom_coords() - gauge_origin
            r_r = np.einsum("ij,ik->ijk", coords, coords)
            res = np.einsum("i,ijk->jk", charges, r_r)
            res = [res[0, 0], res[0, 1], res[0, 2], res[1, 1], res[1, 2], res[2, 2]]
            return np.array(res)
        else:
            raise NotImplementedError("get_nuclear_multipole with order > 2")

    def transform_gauge_origin_to_xyz(self, gauge_origin):
        return _transform_gauge_origin_to_xyz(self.scfres, gauge_origin)

    def fill_occupation_f(self, out):
        if self.restricted:
            out[:] = np.hstack((self.scfres.mo_occ / 2,
                                self.scfres.mo_occ / 2))
        else:
            out[:] = np.hstack((self.scfres.mo_occ[0],
                                self.scfres.mo_occ[1]))

    def fill_orbcoeff_fb(self, out):
        if self.restricted:
            mo_coeff = (self.scfres.mo_coeff,
                        self.scfres.mo_coeff)
        else:
            mo_coeff = self.scfres.mo_coeff
        out[:] = np.transpose(
            np.hstack((mo_coeff[0], mo_coeff[1]))
        )

    def fill_orben_f(self, out):
        if self.restricted:
            out[:] = np.hstack((self.scfres.mo_energy,
                                self.scfres.mo_energy))
        else:
            out[:] = np.hstack((self.scfres.mo_energy[0],
                                self.scfres.mo_energy[1]))

    def fill_fock_ff(self, slices, out):
        diagonal = np.empty(self.n_orbs)
        self.fill_orben_f(diagonal)
        out[:] = np.diag(diagonal)[slices]

    def fill_eri_ffff(self, slices, out):
        self.eri_builder.fill_slice_symm(slices, out)

    def fill_eri_phys_asym_ffff(self, slices, out):
        raise NotImplementedError("fill_eri_phys_asym_ffff not implemented.")

    def has_eri_phys_asym_ffff(self):
        return False

    def flush_cache(self):
        self.eri_builder.flush_cache()


def import_scf(scfres):
    # TODO The error messages here could be a bit more verbose

    if not isinstance(scfres, scf.hf.SCF):
        raise InvalidReference("Unsupported type for backends.pyscf.import_scf.")

    if not scfres.converged:
        raise InvalidReference("Cannot start an adc calculation on top of an SCF,"
                               " which is not yet converged. Did you forget to"
                               " run the kernel() or the scf() function of the"
                               " pyscf scf object?")

    # TODO Check for point-group symmetry,
    #      check for density-fitting or choleski

    return PyScfHFProvider(scfres)


def run_hf(xyz, basis, charge=0, multiplicity=1, conv_tol=1e-11,
           conv_tol_grad=1e-9, max_iter=150, pe_options=None):
    mol = gto.M(
        atom=xyz,
        basis=basis,
        unit="Bohr",
        # spin in the pyscf world is 2S
        spin=multiplicity - 1,
        charge=charge,
        # Disable commandline argument parsing in pyscf
        parse_arg=False,
        dump_input=False,
        verbose=0,
    )
    if pe_options:
        from pyscf.solvent import PE
        mf = PE(scf.HF(mol), pe_options)
    else:
        mf = scf.HF(mol)
    mf.conv_tol = conv_tol
    mf.conv_tol_grad = conv_tol_grad
    mf.max_cycle = max_iter
    # since we want super tight convergence for tests,
    # tweak the options for non-RHF systems
    if multiplicity != 1:
        mf.max_cycle += 500
        mf.diis = scf.EDIIS()
        mf.diis_space = 3
        mf = scf.addons.frac_occ(mf)
    mf.kernel()
    return mf


def run_core_hole(xyz, basis, charge=0, multiplicity=1,
                  conv_tol=1e-11, conv_tol_grad=1e-9, max_iter=150):
    mol = gto.M(
        atom=xyz,
        basis=basis,
        unit="Bohr",
        # spin in the pyscf world is 2S
        spin=multiplicity - 1,
        charge=charge,
        # Disable commandline argument parsing in pyscf
        parse_arg=False,
        dump_input=False,
        verbose=0,
    )

    # First normal run
    mf = scf.UHF(mol)
    mf.conv_tol = conv_tol
    mf.conv_tol_grad = conv_tol_grad
    mf.max_cycle = max_iter
    # since we want super tight convergence for tests,
    # tweak the options for non-RHF systems
    if multiplicity != 1:
        mf.max_cycle += 500
        mf.diis = scf.EDIIS()
        mf.diis_space = 3
        mf = scf.addons.frac_occ(mf)
    mf.kernel()

    # make beta core hole
    mo0 = tuple(c.copy() for c in mf.mo_coeff)
    occ0 = tuple(o.copy() for o in mf.mo_occ)
    occ0[1][0] = 0.0
    dm0 = mf.make_rdm1(mo0, occ0)

    # Run second SCF with MOM
    mf_chole = scf.UHF(mol)
    scf.addons.mom_occ_(mf_chole, mo0, occ0)
    mf_chole.conv_tol = conv_tol
    mf_chole.conv_tol_grad = conv_tol_grad
    mf_chole.max_cycle += 500
    mf_chole.diis = scf.EDIIS()
    mf_chole.diis_space = 3
    mf_chole.kernel(dm0)
    return mf_chole


def _transform_gauge_origin_to_xyz(scfres, gauge_origin):
    """
    Determines the gauge origin. If the gauge origin is defined as a tuple
    the coordinates need to be given in atomic units!
    """
    coords = scfres.mol.atom_coords()
    masses = scfres.mol.atom_mass_list(isotope_avg=True)
    charges = scfres.mol.atom_charges()
    if gauge_origin == "mass_center":
        gauge_origin = tuple(np.einsum("i,ij->j", masses, coords) / masses.sum())
    elif gauge_origin == "charge_center":
        gauge_origin = tuple(np.einsum("i,ij->j", charges, coords)
                             / charges.sum())
    elif gauge_origin == "origin":
        gauge_origin = (0.0, 0.0, 0.0)
    elif isinstance(gauge_origin, tuple):
        gauge_origin = gauge_origin
    else:
        raise NotImplementedError("The gauge origin can be defined either by a "
                                  "keyword (origin, mass_center or charge_center) "
                                  "or by a tuple defining the Cartesian components "
                                  "e.g. (x, y, z)."
                                  )
    return gauge_origin



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2018 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
from libadcc import amplitude_vector_enforce_spin_kind

from adcc import evaluate
from adcc.AmplitudeVector import AmplitudeVector

# TODO
#    This interface is not that great and leads to duplicate information
#    (e.g. once for setting up the guesses and once for setting up the
#     explicit symmetrisation)
#    Maybe one could pass the guesses to these classes or construct
#    these classes along with the guesses or allow to do the guess symmetry
#    setup first and then use this symmetry setup for setting up both the
#    guesses and these classes (which is probably the best case)


class IndexSymmetrisation():
    """
    Enforce the very index symmetrisation required for a particular
    ADC matrix at hand in the new amplitude vectors.
    """
    def __init__(self, matrix):
        # Build symmetrisation functions required to be executed
        # for the respective block
        self.symmetrisation_functions = \
            matrix.construct_symmetrisation_for_blocks()

    def symmetrise(self, new_vectors):
        """
        Symmetrise a set of new vectors to be added to the subspace.

        new_vectors          Vectors to symmetrise (updated in-place)

        Returns:
            The updated new_vectors
        """
        if isinstance(new_vectors, AmplitudeVector):
            return self.symmetrise([new_vectors])[0]
        for vec in new_vectors:
            if not isinstance(vec, AmplitudeVector):
                raise TypeError("new_vectors has to be an "
                                "iterable of AmplitudeVector")
            for b in vec.blocks:
                if b not in self.symmetrisation_functions:
                    continue
                vec[b] = evaluate(self.symmetrisation_functions[b](vec[b]))
        return new_vectors


class IndexSpinSymmetrisation(IndexSymmetrisation):
    """
    Enforce both the required index symmetry as well as an additional
    explicit spin symmetry in the new amplitude vectors.
    """
    def __init__(self, matrix, enforce_spin_kind="singlet"):
        super().__init__(matrix)
        self.enforce_spin_kind = enforce_spin_kind

    def symmetrise(self, new_vectors):
        if isinstance(new_vectors, AmplitudeVector):
            return self.symmetrise([new_vectors])[0]
        new_vectors = super().symmetrise(new_vectors)

        # Enforce singlet (or other spin_kind) spin in the doubles block
        # of all amplitude vectors
        for vec in new_vectors:
            # Only work on the doubles part
            # the other blocks are not yet implemented
            # or nothing needs to be done ("ph" block)
            if "pphh" in vec.blocks:
                # TODO: Note that the "d" is needed here because the C++ side
                #       does not yet understand ph and pphh
                amplitude_vector_enforce_spin_kind(
                    vec.pphh, "d", self.enforce_spin_kind
                )
        return new_vectors


IndexSpinSymmetrisation.symmetrise.__doc__ = \
    IndexSymmetrisation.symmetrise.__doc__



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2018 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
from adcc import block as b
from adcc.LazyMp import LazyMp
from adcc.AdcMethod import IsrMethod
from adcc.functions import einsum
from adcc.Intermediates import Intermediates
from adcc.AmplitudeVector import AmplitudeVector
from adcc.OneParticleDensity import OneParticleDensity
from adcc.NParticleOperator import OperatorSymmetry

from .util import check_doubles_amplitudes, check_singles_amplitudes


def s2s_tdm_isr0(mp, amplitude_l, amplitude_r, intermediates):
    check_singles_amplitudes([b.o, b.v], amplitude_l, amplitude_r)
    ul1 = amplitude_l.ph
    ur1 = amplitude_r.ph

    dm = OneParticleDensity(mp, symmetry=OperatorSymmetry.NOSYMMETRY)
    dm.oo = -einsum('ja,ia->ij', ul1, ur1)
    dm.vv = einsum('ia,ib->ab', ul1, ur1)
    return dm


def s2s_tdm_isr1(mp, amplitude_l, amplitude_r, intermediates):
    dm = s2s_tdm_isr0(mp, amplitude_l, amplitude_r, intermediates)

    try:
        check_doubles_amplitudes([b.o, b.o, b.v, b.v], amplitude_l, amplitude_r)
        ul1, ul2 = amplitude_l.ph, amplitude_l.pphh
        ur1, ur2 = amplitude_r.ph, amplitude_r.pphh
        dm.ov += -2.0 * einsum("jb,ijab->ia", ul1, ur2)
        dm.vo += -2.0 * einsum("ijab,jb->ai", ul2, ur1)
    except ValueError:
        pass

    return dm


def s2s_tdm_isr2(mp, amplitude_l, amplitude_r, intermediates):
    check_doubles_amplitudes([b.o, b.o, b.v, b.v], amplitude_l, amplitude_r)
    dm = s2s_tdm_isr1(mp, amplitude_l, amplitude_r, intermediates)

    ul1, ul2 = amplitude_l.ph, amplitude_l.pphh
    ur1, ur2 = amplitude_r.ph, amplitude_r.pphh

    t2 = mp.t2(b.oovv)
    p0 = mp.mp2_diffdm
    p1_oo = dm.oo.evaluate()  # ADC(1) tdm
    p1_vv = dm.vv.evaluate()  # ADC(1) tdm

    # ADC(2) ISR intermediate (TODO Move to intermediates)
    rul1 = einsum('ijab,jb->ia', t2, ul1).evaluate()
    rur1 = einsum('ijab,jb->ia', t2, ur1).evaluate()

    dm.oo += (
        - 2.0 * einsum('ikab,jkab->ij', ur2, ul2)
        + 0.5 * einsum('ik,kj->ij', p1_oo, p0.oo)
        + 0.5 * einsum('ik,kj->ij', p0.oo, p1_oo)
        - 0.5 * einsum('ikcd,lk,jlcd->ij', t2, p1_oo, t2)
        + 1.0 * einsum('ikcd,jkcb,db->ij', t2, t2, p1_vv)
        - 0.5 * einsum('ia,jkac,kc->ij', ur1, t2, rul1)
        - 0.5 * einsum('ikac,kc,ja->ij', t2, rur1, ul1)
        - 1.0 * einsum('ia,ja->ij', rul1, rur1)
    )
    dm.vv += (
        + 2.0 * einsum('ijac,ijbc->ab', ul2, ur2)
        - 0.5 * einsum("ac,cb->ab", p1_vv, p0.vv)
        - 0.5 * einsum("ac,cb->ab", p0.vv, p1_vv)
        - 0.5 * einsum("klbc,klad,cd->ab", t2, t2, p1_vv)
        + 1.0 * einsum("klbc,jk,jlac->ab", t2, p1_oo, t2)
        + 0.5 * einsum("ikac,kc,ib->ab", t2, rul1, ur1)
        + 0.5 * einsum("ia,ikbc,kc->ab", ul1, t2, rur1)
        + 1.0 * einsum("ia,ib->ab", rur1, rul1)
    )

    # (TODO Move to intermediates)
    p1_ov = -2.0 * einsum("jb,ijab->ia", ul1, ur2).evaluate()
    p1_vo = -2.0 * einsum("ijab,jb->ai", ul2, ur1).evaluate()

    dm.ov += (
        - einsum("ijab,bj->ia", t2, p1_vo)
        - einsum("ib,ba->ia", p0.ov, p1_vv)
        + einsum("ij,ja->ia", p1_oo, p0.ov)
        - einsum("ib,klca,klcb->ia", ur1, t2, ul2)
        - einsum("ikcd,jkcd,ja->ia", t2, ul2, ur1)
    )
    dm.vo += (
        - einsum("ijab,jb->ai", t2, p1_ov)
        - einsum("ib,ab->ai", p0.ov, p1_vv)
        + einsum("ji,ja->ai", p1_oo, p0.ov)
        - einsum("ib,klca,klcb->ai", ul1, t2, ur2)
        - einsum("ikcd,jkcd,ja->ai", t2, ur2, ul1)
    )
    return dm


# Ref: https://doi.org/10.1080/00268976.2013.859313
DISPATCH = {
    "isr0": s2s_tdm_isr0,
    "isr1s": s2s_tdm_isr0,  # Identical to ISR(0)
    "isr1": s2s_tdm_isr1,
    "isr2d": s2s_tdm_isr2,  # Identical to ISR(2)
    "isr2": s2s_tdm_isr2,
}


def state2state_transition_dm(method, ground_state, amplitude_from,
                              amplitude_to, intermediates=None):
    """
    Compute the state to state transition density matrix
    state in the MO basis using the intermediate-states representation.

    Parameters
    ----------
    method : str, IsrMethod
        The method to use for the computation (e.g. "isr2")
    ground_state : LazyMp
        The ground state upon which the excitation was based
    amplitude_from : AmplitudeVector
        The amplitude vector of the state to start from
    amplitude_to : AmplitudeVector
        The amplitude vector of the state to excite to
    intermediates : adcc.Intermediates
        Intermediates from the ADC calculation to reuse
    """
    if not isinstance(method, IsrMethod):
        method = IsrMethod(method)
    if not isinstance(ground_state, LazyMp):
        raise TypeError("ground_state should be a LazyMp object.")
    if not isinstance(amplitude_from, AmplitudeVector):
        raise TypeError("amplitude_from should be an AmplitudeVector object.")
    if not isinstance(amplitude_to, AmplitudeVector):
        raise TypeError("amplitude_to should be an AmplitudeVector object.")
    if intermediates is None:
        intermediates = Intermediates(ground_state)

    if method.name not in DISPATCH:
        raise NotImplementedError("state2state_transition_dm is not implemented "
                                  f"for {method.name}.")
    else:
        # final state is on the bra side/left (complex conjugate)
        # see ref https://doi.org/10.1080/00268976.2013.859313, appendix A2
        ret = DISPATCH[method.name](ground_state, amplitude_to, amplitude_from,
                                    intermediates)
        return ret.evaluate()



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2019 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
import os
import subprocess


def build_docs():
    subprocess.run("rm -r docs/api".split())
    subprocess.run("./setup.py build_docs".split(), check=True)
    return "build/sphinx/html"


def upload_docs(outdir):
    import json

    with open(os.path.dirname(__file__) + "/config.json") as fp:
        target = json.load(fp)["documentation"]
    subprocess.run("rsync -P -rvzc --exclude .buildinfo --exclude objects.inv "
                   "--delete {}/ {} --cvs-exclude".format(outdir, target).split(),
                   check=True)


def main():
    if not os.path.isfile("scripts/upload_documentation.py") or \
       not os.path.isfile("setup.py"):
        raise SystemExit("Please run from top dir of repository")

    outdir = build_docs()
    upload_docs(outdir)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2019 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
import os
import tempfile
import warnings
import numpy as np

from mpi4py import MPI
from libadcc import HartreeFockProvider

import veloxchem as vlx

from .EriBuilder import EriBuilder
from ..exceptions import InvalidReference
from ..ElectronicStates import EnergyCorrection

from veloxchem.mpitask import MpiTask
from veloxchem.veloxchemlib import (AngularMomentumIntegralsDriver,
                                    LinearMomentumIntegralsDriver)


class VeloxChemOperatorIntegralProvider:
    available: tuple[str, ...] = (
        "electric_dipole", "electric_dipole_velocity", "magnetic_dipole",
    )

    def __init__(self, scfdrv):
        self.scfdrv = scfdrv
        self.backend = "veloxchem"

    @property
    def electric_dipole(self) -> tuple[np.ndarray, ...]:
        """-sum_i r_i"""
        task = self.scfdrv.task
        dipole_drv = vlx.ElectricDipoleIntegralsDriver(task.mpi_comm)
        # define the origin for electric dipole integrals
        dipole_drv.origin = tuple(np.zeros(3))
        dipole_mats = dipole_drv.compute(task.molecule, task.ao_basis)
        return (-1.0 * dipole_mats.x_to_numpy(),
                -1.0 * dipole_mats.y_to_numpy(),
                -1.0 * dipole_mats.z_to_numpy())

    def magnetic_dipole(self, gauge_origin="origin") -> tuple[np.ndarray, ...]:
        """
        The imaginary part of the integral is returned.
        -0.5 * sum_i r_i x p_i
        """
        gauge_origin = _transform_gauge_origin_to_xyz(self.scfdrv, gauge_origin)
        task = self.scfdrv.task
        angmom_drv = AngularMomentumIntegralsDriver(task.mpi_comm)
        angmom_drv.origin = tuple(gauge_origin)
        angmom_mats = angmom_drv.compute(task.molecule, task.ao_basis)
        return (0.5 * angmom_mats.x_to_numpy(),
                0.5 * angmom_mats.y_to_numpy(),
                0.5 * angmom_mats.z_to_numpy())

    @property
    def electric_dipole_velocity(self) -> tuple[np.ndarray, ...]:
        """
        The imaginary part of the integral is returned.
        -sum_i p_i
        """
        task = self.scfdrv.task
        linmom_drv = LinearMomentumIntegralsDriver(task.mpi_comm)
        linmom_mats = linmom_drv.compute(task.molecule, task.ao_basis)
        return (-1.0 * linmom_mats.x_to_numpy(),
                -1.0 * linmom_mats.y_to_numpy(),
                -1.0 * linmom_mats.z_to_numpy())


class VeloxChemEriBuilder(EriBuilder):
    def __init__(self, task, mol_orbs, n_orbs, n_orbs_alpha, n_alpha,
                 n_beta, restricted):
        self.moints_drv = vlx.MOIntegralsDriver(task.mpi_comm, task.ostream)
        self.compute_args = (task.molecule, task.ao_basis, mol_orbs)
        super().__init__(n_orbs, n_orbs_alpha, n_alpha, n_beta, restricted)

    def compute_mo_eri(self, blocks, spins):
        eri = self.moints_drv.compute_in_memory(*self.compute_args,
                                                moints_name="chem_" + blocks,
                                                moints_spin=spins)
        return eri


class VeloxChemHFProvider(HartreeFockProvider):
    """
        This implementation is valid for RHF and UHF
    """
    def __init__(self, scfdrv):
        # Do not forget the next line,
        # otherwise weird errors result
        super().__init__()
        warnings.warn(
            "The VeloxChem interface does not work with new VeloxChem versions "
            "and is no longer maintained. It will be removed in the future.",
            DeprecationWarning
        )

        self.scfdrv = scfdrv
        self.mol_orbs = self.scfdrv.mol_orbs
        self.molecule = self.scfdrv.task.molecule
        self.ao_basis = self.scfdrv.task.ao_basis
        self.mpi_comm = self.scfdrv.task.mpi_comm
        self.ostream = self.scfdrv.task.ostream
        n_alpha = self.molecule.number_of_alpha_electrons()
        n_beta = self.molecule.number_of_beta_electrons()
        self.eri_builder = VeloxChemEriBuilder(
            self.scfdrv.task, self.mol_orbs, self.n_orbs, self.n_orbs_alpha,
            n_alpha, n_beta, self.restricted
        )

        self.operator_integral_provider = VeloxChemOperatorIntegralProvider(
            self.scfdrv
        )

    def pe_energy(self, dm, elec_only=True):
        e_pe, _ = self.scfdrv.pe_drv.get_pe_contribution(dm.to_ndarray(),
                                                         elec_only=elec_only)
        return e_pe

    @property
    def excitation_energy_corrections(self):
        ret = []
        if hasattr(self.scfdrv, "pe_drv"):
            ptlr = EnergyCorrection(
                "pe_ptlr_correction",
                lambda view: 2.0 * self.pe_energy(view.transition_dm_ao,
                                                  elec_only=True)
            )
            ptss = EnergyCorrection(
                "pe_ptss_correction",
                lambda view: self.pe_energy(view.state_diffdm_ao,
                                            elec_only=True)
            )
            ret.extend([ptlr, ptss])
        return {ec.name: ec for ec in ret}

    def get_backend(self):
        return "veloxchem"

    def get_conv_tol(self):
        return self.scfdrv.conv_thresh

    def get_restricted(self):
        if isinstance(self.scfdrv, vlx.scfrestdriver.ScfRestrictedDriver):
            restricted = True
        else:
            restricted = False
        return restricted

    def get_energy_scf(self):
        return self.scfdrv.get_scf_energy()

    def get_spin_multiplicity(self):
        return self.molecule.get_multiplicity()

    def get_n_orbs_alpha(self):
        return self.mol_orbs.alpha_to_numpy().shape[1]

    def get_n_bas(self):
        return self.mol_orbs.number_aos()

    def get_nuclear_multipole(self, order, gauge_origin=(0, 0, 0)):
        mol = self.scfdrv.task.molecule
        nuc_charges = mol.elem_ids_to_numpy()
        if order == 0:
            # The function interface needs to be a np.array on return
            return np.array([np.sum(nuc_charges)])
        elif order == 1:
            coords = np.vstack(
                (mol.x_to_numpy(), mol.y_to_numpy(), mol.z_to_numpy())
            ).transpose()
            return np.einsum('i,ix->x', nuc_charges, coords)
        elif order == 2:
            coords = mol.get_coordinates_in_bohr() - gauge_origin
            r_r = np.einsum("ij,ik->ijk", coords, coords)
            res = np.einsum("i,ijk->jk", nuc_charges, r_r)
            res = [res[0, 0], res[0, 1], res[0, 2], res[1, 1], res[1, 2], res[2, 2]]
            return np.array(res)
        else:
            raise NotImplementedError("get_nuclear_multipole with order > 2")

    def transform_gauge_origin_to_xyz(self, gauge_origin):
        return _transform_gauge_origin_to_xyz(self.scfdrv, gauge_origin)

    def fill_occupation_f(self, out):
        n_mo = self.mol_orbs.number_mos()
        occ_a = self.molecule.get_aufbau_alpha_occupation(n_mo)
        occ_b = self.molecule.get_aufbau_beta_occupation(n_mo)
        out[:] = np.hstack((occ_a, occ_b))

    def fill_orbcoeff_fb(self, out):
        mo_coeff_a = self.mol_orbs.alpha_to_numpy()
        mo_coeff_b = self.mol_orbs.beta_to_numpy()
        out[:] = np.transpose(
            np.hstack((mo_coeff_a, mo_coeff_b))
        )

    def fill_orben_f(self, out):
        orben_a = self.mol_orbs.ea_to_numpy()
        orben_b = self.mol_orbs.eb_to_numpy()
        out[:] = np.hstack((orben_a, orben_b))

    def fill_fock_ff(self, slices, out):
        diagonal = np.empty(self.n_orbs)
        self.fill_orben_f(diagonal)
        out[:] = np.diag(diagonal)[slices]

    def fill_eri_ffff(self, slices, out):
        self.eri_builder.fill_slice_symm(slices, out)

    def fill_eri_phys_asym_ffff(self, slices, out):
        raise NotImplementedError("fill_eri_phys_asym_ffff not implemented.")

    def has_eri_phys_asym_ffff(self):
        return False

    def flush_cache(self):
        self.eri_builder.flush_cache()


def import_scf(scfdrv):
    # TODO The error messages in here could be a little more informative

    if not hasattr(scfdrv, "task"):
        raise InvalidReference("Please attach the VeloxChem task to "
                               "the VeloxChem SCF driver")

    if not scfdrv.is_converged:
        raise InvalidReference("Cannot start an adc calculation on top "
                               "of an SCF, which is not converged.")

    provider = VeloxChemHFProvider(scfdrv)
    return provider


def run_hf(xyz, basis, charge=0, multiplicity=1, conv_tol=None, conv_tol_grad=1e-9,
           max_iter=150, pe_options=None):
    basis_remap = {
        "sto3g": "sto-3g",
        "def2tzvp": "def2-tzvp",
        "ccpvdz": "cc-pvdz",
    }
    with tempfile.TemporaryDirectory() as tmpdir:
        infile = os.path.join(tmpdir, "vlx.in")
        outfile = os.path.join(tmpdir, "vlx.out")
        with open(infile, "w") as fp:
            lines = ["@jobs", "task: hf", "@end", ""]
            lines += ["@scf",
                      "conv_thresh: {}".format(conv_tol_grad),
                      "max_iter: {}".format(max_iter),
                      "@end", ""]
            lines += ["@method settings",
                      "basis: {}".format(basis_remap.get(basis, basis))]
            # TODO: PE results in VeloxChem are currently wrong, because
            # polarizabilities are always made isotropic
            if pe_options:
                potfile = pe_options["potfile"]
                lines += ["pe: yes",
                          f"potfile: {potfile}"]
            lines += ["@end"]
            lines += ["@molecule",
                      "charge: {}".format(charge),
                      "multiplicity: {}".format(multiplicity),
                      "units: bohr",
                      "xyz:\n{}".format("\n".join(xyz.split(";"))),
                      "@end"]
            fp.write("\n".join(lines))
        task = MpiTask([infile, outfile], MPI.COMM_WORLD)
        if multiplicity == 1:
            scfdrv = vlx.ScfRestrictedDriver(task.mpi_comm, task.ostream)
        else:
            scfdrv = vlx.ScfUnrestrictedDriver(task.mpi_comm, task.ostream)
        scfdrv.update_settings(task.input_dict['scf'],
                               task.input_dict['method_settings'])
        scfdrv.compute(task.molecule, task.ao_basis, task.min_basis)
        scfdrv.task = task
    return scfdrv


def _transform_gauge_origin_to_xyz(scfdrv, gauge_origin):
    """
    Determines the gauge origin. If the gauge origin is defined as a tuple
    the coordinates need to be given in atomic units!
    """
    molecule = scfdrv.task.molecule
    coords = molecule.get_coordinates_in_bohr()
    charges = molecule.elem_ids_to_numpy()
    masses = molecule.masses_to_numpy()

    if gauge_origin == "mass_center":
        gauge_origin = tuple(np.einsum("i,ij->j", masses, coords) / masses.sum())
    elif gauge_origin == "charge_center":
        gauge_origin = tuple(np.einsum("i,ij->j", charges, coords)
                             / charges.sum())
    elif gauge_origin == "origin":
        gauge_origin = (0.0, 0.0, 0.0)
    elif isinstance(gauge_origin, tuple):
        gauge_origin = gauge_origin
    else:
        raise NotImplementedError("The gauge origin can be defined either by a "
                                  "keyword (origin, mass_center or charge_center) "
                                  "or by a tuple defining the Cartesian components "
                                  "e.g. (x, y, z)."
                                  )
    return gauge_origin



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2019 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
import numpy as np

from . import shapefctns
from ..misc import requires_module


class Spectrum:
    """
    Class for representing arbitrary spectra. This design is strongly
    inspired by the approach used by pymatgen.
    """

    # TODO Ideas:
    #   - normalise()   normalise spectrum by scaling it
    #   - shift()       shift spectrum such that max peak is at a
    #                   particular posn
    #   - interpolate_value(self, x, order=1):
    #                   interpolate value by linear / polynomial interpolation
    #                   between existing points
    #  support element-wise multiplication, division, addition /
    #                   subtraction of spectra ?

    def __init__(self, x, y, *args, xlabel: str = None, ylabel: str = None,
                 **kwargs):
        """Pass spectrum data to initialise the class.

        To allow the copy and other functions to work properly, all arguments
        and keywords arguments from implementing classes should be passed here.

        Parameters
        ----------
        x : ndarray
            Data on the first axis
        y : ndarray
            Data on the second axis
        args
            Arguments passed to subclass upon construction
        kwargs
            Keyword arguments passed to subclass upon construction.
        """
        self.x = np.array(x).flatten()
        self.y = np.array(y).flatten()
        self.xlabel = "x" if xlabel is None else xlabel
        self.ylabel = "y" if ylabel is None else ylabel
        if self.x.size != self.y.size:
            raise ValueError("Sizes of x and y mismatch: {} versus {}."
                             "".format(self.x.size, self.y.size))
        self._args = args
        self._kwargs = kwargs

    def broaden_lines(self, width=None, shape="lorentzian", xmin=None, xmax=None):
        """Apply broadening to the current spectral data and
        return the broadened spectrum.

        Parameters
        ----------
        shape : str or callable, optional
            The shape of the broadening to use (lorentzian or gaussian),
            by default lorentzian broadening is used. This can be a callable
            to directly specify the function with which each line of the
            spectrum is convoluted.
        width : float, optional
            The width to use for the broadening (stddev for the gaussian,
            gamma parameter for the lorentzian).
            Optional if shape is a callable.
        xmin : float, optional
            Explicitly set the minimum value of the x-axis for broadening
        xmax : float, optional
            Explicitly set the maximum value of the x-axis for broadening
        """
        if not callable(shape) and width is None:
            raise ValueError("If shape is not a callable, the width parameter "
                             "is required")

        if not callable(shape):
            if not hasattr(shapefctns, shape):
                raise ValueError("Unknown broadening function: " + shape)
            shapefctn = getattr(shapefctns, shape)

            def shape(x, x0):
                # Empirical scaling factor to make the envelope look nice
                scale = width / 0.272
                return scale * shapefctn(x, x0, width)

        # the min and max of the underlying spectrum
        xmin_disc = np.min(self.x)
        xmax_disc = np.max(self.x)
        # determine a padding that is added to the left and right of the lowest
        # and highest point of the discrete spectrum, respectively.
        # But only if the user did not provide any data for min/max!
        if xmin is None:
            xmin = xmin_disc
            xmin_padding = max((xmax_disc - xmin_disc) / 10, 5 * width)
        else:
            xmin_padding = 0
        if xmax is None:
            xmax = xmax_disc
            xmax_padding = max((xmax_disc - xmin_disc) / 10, 5 * width)
        else:
            xmax_padding = 0

        n_points = min(5000, max(500, int(1000 * (xmax - xmin))))
        x = np.linspace(xmin - xmin_padding, xmax + xmax_padding, n_points)

        y = 0
        for center, value in zip(self.x, self.y):
            y += value * shape(x, center)

        cpy = self.copy()
        cpy.x = x
        cpy.y = y
        return cpy

    def copy(self):
        """Return a consistent copy of the object."""
        return self.__class__(
            self.x, self.y, *self._args, xlabel=self.xlabel, ylabel=self.ylabel,
            **self._kwargs
        )

    @requires_module("matplotlib")
    def plot(self, *args, style=None, **kwargs):
        """Plot the Spectrum represented by this class.

        Parameters not listed below are passed to the matplotlib plot function.

        Parameters
        ----------
        style : str, optional
            Use some default setup of matplotlib parameters for certain
            types of spectra commonly plotted. Valid are "discrete" and
            "continuous". By default no special style is chosen.
        """
        from matplotlib import pyplot as plt
        if style == "discrete":
            p = plt.plot(self.x, self.y, "x", *args, **kwargs)
            plt.vlines(self.x, 0, self.y, linestyle="dashed",
                       color=p[0].get_color(), linewidth=1)
        elif style == "continuous":
            p = plt.plot(self.x, self.y, "-", *args, **kwargs)
        elif style is None:
            p = plt.plot(self.x, self.y, *args, **kwargs)
        else:
            raise ValueError("Unknown style: " + str(style))
        plt.xlabel(self.xlabel)
        plt.ylabel(self.ylabel)
        # if we have negative y-values (e.g., rotatory strengths),
        # draw y = 0 as an extra line for clarity
        if np.any(self.y < 0.0):
            plt.axhline(0.0, color='gray', lw=0.5)
        return p



#!/usr/bin/env python3

import adcc
from pyscf import gto, scf
from pyscf.solvent import ddCOSMO

# Run PCM SCF in pyscf
mol = gto.M(
    atom="""
    C          8.64800        1.07500       -1.71100
    C          9.48200        0.43000       -0.80800
    C          9.39600        0.75000        0.53800
    C          8.48200        1.71200        0.99500
    C          7.65300        2.34500        0.05500
    C          7.73200        2.03100       -1.29200
    H         10.18300       -0.30900       -1.16400
    H         10.04400        0.25200        1.24700
    H          6.94200        3.08900        0.38900
    H          7.09700        2.51500       -2.01800
    N          8.40100        2.02500        2.32500
    N          8.73400        0.74100       -3.12900
    O          7.98000        1.33100       -3.90100
    O          9.55600       -0.11000       -3.46600
    H          7.74900        2.71100        2.65200
    H          8.99100        1.57500        2.99500
    """,
    basis='sto-3g', symmetry=0, charge=0, spin=0,
    unit="Angström"
)

mf = ddCOSMO(scf.RHF(mol))
# set the dielectric constant
mf.with_solvent.eps = 78.36
mf.conv_tol = 1e-8
mf.conv_tol_grad = 1e-7
mf.max_cycle = 150

mf.kernel()

# Run ADC2 with with linear-response for the solvent

# first the dielectric constant needs to be adjusted to
# the corresponding optical dielectric constant.
# Note that this is also necessary for the ptlr scheme.
mf.with_solvent.eps = 1.78
state = adcc.adc2(mf, n_singlets=5, conv_tol=1e-6,
                  environment="linear_response")
print(state.describe())



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2018 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
import numpy as np

from molsturm.State import State
from adcc.DataHfProvider import DataHfProvider


def convert_scf_to_dict(scfres):
    if not isinstance(scfres, State):
        raise TypeError("Unsupported type for backends.molsturm.import_scf.")

    n_alpha = scfres["n_alpha"]
    n_beta = scfres["n_beta"]
    data = {}
    if scfres["n_orbs_alpha"] != scfres["n_orbs_beta"]:
        raise ValueError("n_orbs_alpha != n_orbs_beta not supported.")

    # Keys to include verbatim from the hf res dictionary to the adcc input
    verbatim_keys = [
        "n_orbs_alpha", "restricted", "orben_f", "eri_ffff", "fock_ff",
    ]
    for k in verbatim_keys:
        data[k] = scfres[k]

    if scfres["restricted"]:
        data["spin_multiplicity"] = 2 * (n_alpha - n_beta) + 1
    else:
        data["spin_multiplicity"] = 0

    n_oa = data["n_orbs_alpha"]
    data["occupation_f"] = np.zeros(2 * n_oa)
    data["occupation_f"][:n_alpha] = 1.
    data["occupation_f"][n_oa:n_oa + n_beta] = 1.

    data["energy_scf"] = scfres["energy_ground_state"]
    data["conv_tol"] = scfres["final_error_norm"]
    data["orbcoeff_fb"] = scfres["orbcoeff_bf"].transpose().copy()

    # Compute electric and nuclear multipole moments
    data["multipoles"] = {"elec_0": -int(n_alpha + n_beta), }
    if "input_parameters" in scfres:
        params = scfres["input_parameters"]
        coords = np.asarray(params["system"]["coords"])
        charges = np.asarray(params["system"]["atom_numbers"])
        data["multipoles"]["nuclear_0"] = int(np.sum(charges)),
        data["multipoles"]["nuclear_1"] = np.einsum('i,ix->x', charges, coords)
    else:
        import warnings

        # We have no information about this, so we can just provide dummies
        data["multipoles"]["nuclear_0"] = -1
        data["multipoles"]["nuclear_1"] = np.zeros(3)
        warnings.warn("The passed molsturm scfres has no information about "
                      "nuclear multipoles, such that dummy data are used. "
                      "Results involving nuclear multipoles could be wrong.")

    data["backend"] = "molsturm"
    return data


def import_scf(scfres):
    return DataHfProvider(convert_scf_to_dict(scfres))


basis_remap = {
    "sto3g": "sto-3g",
    "def2tzvp": "def2-tzvp",
    "ccpvdz": "cc-pvdz",
}


def run_hf(xyz, basis, charge=0, multiplicity=1, conv_tol=1e-12,
           conv_tol_grad=1e-8, max_iter=150):

    import molsturm

    # Quick-and-dirty xyz parser:
    geom = xyz.split()
    n_atom = len(geom) // 4
    assert n_atom * 4 == len(geom)
    atoms = [geom[i * 4] for i in range(n_atom)]
    coords = [[float(geom[i * 4 + 1]),
               float(geom[i * 4 + 2]),
               float(geom[i * 4 + 3])] for i in range(n_atom)]

    mol = molsturm.System(atoms, coords)
    mol.charge = charge
    mol.multiplicity = multiplicity

    return molsturm.hartree_fock(mol, basis_type="gaussian",
                                 basis_set_name=basis_remap.get(basis, basis),
                                 conv_tol=conv_tol_grad, max_iter=max_iter)



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
import adcc

from pyscf import gto, scf

# Run SCF in pyscf
mol = gto.M(
    atom='O 0 0 0;'
         'H 0 0 1.795239827225189;'
         'H 1.693194615993441 0 -0.599043184453037',
    basis='sto-3g',
    spin=2,
    unit="Bohr"
)
scfres = scf.UHF(mol)
scfres.conv_tol = 1e-8
scfres.conv_tol_grad = 1e-8
scfres.max_cycle = 100
scfres.verbose = 4
scfres.kernel()

# Run an adc2 calculation:
singlets = adcc.adc2(scfres, n_states=5)

print(singlets.describe())



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
import adcc

from pyscf import gto, scf
from matplotlib import pyplot as plt

# Run SCF in pyscf
mol = gto.M(
    atom='S  -0.38539679062   0 -0.27282082253;'
         'H  -0.0074283962687 0  2.2149138578;'
         'H   2.0860198029    0 -0.74589639249',
    basis='cc-pvtz',
    unit="Bohr"
)
scfres = scf.RHF(mol)
scfres.conv_tol = 1e-13
scfres.kernel()
nmo = mol.nao

scfres.analyze()

print()
print("=========== CVS (1s to 2s) ================")
print()
singlets = adcc.cvs_adc2x(scfres, core_orbitals=2, n_singlets=10)
print(singlets.describe())
singlets.plot_spectrum(label="H2S CVS (1s and 2s)")


print()
print("=========== CVS (2s) ================")
print()
singlets = adcc.cvs_adc2x(scfres, core_orbitals=[1, 1 + nmo], n_singlets=10)
print(singlets.describe())
singlets.plot_spectrum(label="H2S CVS (2s)")

plt.legend()
plt.savefig("pyscf_ccpvtz_arbitrary_cvs_adc2x.pdf")



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2026 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
from dataclasses import dataclass
from enum import Enum
from itertools import product
import math
import numpy as np

import libadcc

from .functions import evaluate
from .MoSpaces import split_spaces
from .Tensor import Tensor


class OperatorSymmetry(Enum):
    NOSYMMETRY = 0
    HERMITIAN = 1
    ANTIHERMITIAN = 2

    def to_str(self) -> str:
        return self.name.lower()


@dataclass(frozen=True)
class BlockInfo:
    canonical: str
    factor: int
    transpose: tuple[int, ...]


class CanonicalBlock:
    def __init__(self):
        self._symmetry_factor = 0
        self._tensor = None

    @property
    def symmetry_factor(self):
        return self._symmetry_factor

    @property
    def tensor(self):
        return self._tensor

    @tensor.setter
    def tensor(self, tensor):
        self._tensor = tensor


class NParticleOperator:
    def __init__(self, spaces, n_particle_op, symmetry=OperatorSymmetry.HERMITIAN):
        """
        Construct an NParticleOperator object. All blocks are initialised
        as zero blocks.

        Parameters
        ----------
        spaces : adcc.MoSpaces or adcc.ReferenceState or adcc.LazyMp
            MoSpaces object

        n_particle_operator: int
            Particle rank of the operator, i.e. the number of creation and
            annihilation operators involved.

        symmetry : OperatorSymmetry, optional
            Symmetry type of the operator. Can be:
            - `OperatorSymmetry.NOSYMMETRY` : No symmetry is enforced
            - `OperatorSymmetry.HERMITIAN` : Operator is Hermitian (O^\\dagger = O)
            - `OperatorSymmetry.ANTIHERMITIAN` : Operator is Antihermitian
                                                 (O^\\dagger = -O)
            Default is `OperatorSymmetry.HERMITIAN`.
        """
        from .block import get_canonical_block

        self._n_particle_op = n_particle_op
        if hasattr(spaces, "mospaces"):
            self.mospaces = spaces.mospaces
        else:
            self.mospaces = spaces
        self.symmetry = symmetry

        self.reference_state = None
        # Set reference_state if possible
        if isinstance(spaces, libadcc.ReferenceState):
            self.reference_state = spaces
        elif hasattr(spaces, "reference_state"):
            assert isinstance(spaces.reference_state, libadcc.ReferenceState)
            self.reference_state = spaces.reference_state

        occs = sorted(self.mospaces.subspaces_occupied)
        virts = sorted(self.mospaces.subspaces_virtual)
        self.orbital_subspaces = (*occs, *virts)

        # check that orbital subspaces are correct
        assert sum(self.mospaces.n_orbs(ss) for ss in self.orbital_subspaces) \
            == self.mospaces.n_orbs("f")

        # Initialize all blocks; symmetry rules are applied lazily upon access.
        combs = list(product(self.orbital_subspaces,
                             repeat=2 * self._n_particle_op))
        blocks = tuple("".join((comb)) for comb in combs)

        self._block_info: dict[str, BlockInfo] = {}
        self._canonical_blocks: dict[str, CanonicalBlock] = {}

        for block in blocks:
            bra = block[:2 * self.n_particle_op]
            ket = block[2 * self.n_particle_op:]
            canonical_block, factor, transpose = get_canonical_block(
                bra, ket, self.symmetry
            )
            self._block_info[block] = BlockInfo(
                canonical=canonical_block,
                factor=factor,
                transpose=transpose,
            )

            # add new canonical block to dictionary
            if canonical_block not in self._canonical_blocks.keys():
                self._canonical_blocks[canonical_block] = CanonicalBlock()

            # increment factor counter
            self._canonical_blocks[canonical_block]._symmetry_factor += 1

    @property
    def canonical_factors(self) -> dict[str, int]:
        """
        Returns canonical block multiplicity factors.
        """
        return {
            cb: value.symmetry_factor
            for cb, value in self._canonical_blocks.items()
        }

    @property
    def canonical_blocks(self) -> tuple[str, ...]:
        """
        Returns tuple of canonical block labels.
        """
        return tuple(self._canonical_blocks.keys())

    @property
    def blocks(self) -> tuple[str, ...]:
        """
        Returns tuple of all block labels.
        """
        return tuple(self._block_info.keys())

    @property
    def n_particle_op(self) -> int:
        """
        Returns the particle rank of the operator.
        """
        return self._n_particle_op

    @property
    def shape(self) -> tuple[int, ...]:
        """
        Returns the shape tuple of the NParticleOperator
        """
        size = self.mospaces.n_orbs("f")
        return (size, size,) * self.n_particle_op

    @property
    def size(self) -> int:
        """
        Returns the number of elements of the NParticleOperator
        """
        return np.prod(self.shape)

    @property
    def blocks_nonzero(self) -> list[str, ...]:
        """
        Returns a list of the non-zero block labels
        """
        return [cb for cb, value
                in self._canonical_blocks.items() if value.tensor is not None]

    def is_zero_block(self, block):
        """
        Checks if block is explicitly marked as zero block.
        Returns False if the block does not exist.
        """
        if block not in self.canonical_blocks:
            return False
        return block not in self.blocks_nonzero

    def block(self, block):
        """
        Returns tensor of the given block.
        Does not create a block in case it is marked as a zero block.
        Use __getitem__ for that purpose.
        """
        block_info = self._block_info.get(block, None)
        if block_info is None:
            raise KeyError(f"Invalid block {block} requested. "
                           f"Available blocks are: {self.blocks}.")

        canonical_block = self._canonical_blocks[block_info.canonical]
        if canonical_block.tensor is None:
            raise KeyError("The block function does not support "
                           "access to zero-blocks. Available non-zero "
                           f"blocks are: {self.blocks_nonzero}.")

        if block == canonical_block:
            return canonical_block.tensor

        return (
            block_info.factor
            * canonical_block.tensor.transpose(block_info.transpose)
        )

    def __getitem__(self, block) -> Tensor:
        if block not in self.blocks:
            raise KeyError(f"Invalid block {block} requested. "
                           f"Available blocks are: {self.blocks}.")

        block_info = self._block_info[block]
        canonical_block = self._canonical_blocks[block_info.canonical]

        if canonical_block.tensor is None:
            sym = libadcc.make_symmetry_operator(
                self.mospaces, block_info.canonical, self.symmetry.to_str(), "1"
            )
            canonical_block.tensor = Tensor(sym)

        if block == block_info.canonical:
            return canonical_block.tensor

        return (
            block_info.factor
            * canonical_block.tensor.transpose(block_info.transpose)
        )

    def __getattr__(self, attr) -> Tensor:
        from . import block as b
        return self.__getitem__(b.__getattr__(attr))

    def __setitem__(self, block, tensor):
        block_info = self._block_info.get(block, None)
        if block_info is None or block_info.canonical != block:
            raise KeyError(f"Invalid block {block} assigned. "
                           f"Available blocks are: {self.canonical_blocks}.")

        spaces = split_spaces(block)
        expected_shape = tuple(self.mospaces.n_orbs(space) for space in spaces)
        if expected_shape != tensor.shape:
            raise ValueError("Invalid shape of incoming tensor. "
                             f"Expected shape {expected_shape}, but "
                             f"got shape {tensor.shape} instead.")
        canonical_block = self._canonical_blocks[block_info.canonical]
        canonical_block.tensor = tensor

    def __setattr__(self, attr, value):
        try:
            from . import block as b
            self.__setitem__(b.__getattr__(attr), value)
        except AttributeError:
            super().__setattr__(attr, value)

    def set_zero_block(self, block):
        """
        Set a given block as zero block
        """
        if block not in self.canonical_blocks:
            raise KeyError(f"Invalid block {block} set as zero block. "
                           f"Available blocks are: {self.blocks}.")
        block_info = self._block_info[block]
        canonical_block = self._canonical_blocks[block_info.canonical]
        canonical_block.tensor = None

    def to_ndarray(self) -> np.ndarray:
        """
        Returns the NParticleOperator as a contiguous
        np.ndarray instance including all blocks
        """
        # offsets to start index of spaces
        offsets = {
            sp: sum(
                self.mospaces.n_orbs(ss)
                for ss in self.orbital_subspaces[:self.orbital_subspaces.index(sp)]
            )
            for sp in self.orbital_subspaces
        }
        # slices for each space
        slices = {
            sp: slice(offsets[sp], offsets[sp] + self.mospaces.n_orbs(sp))
            for sp in self.orbital_subspaces
        }
        # build complete np.ndarray -> go through all blocks
        ret = np.zeros((self.shape))
        for block in self.blocks:
            if self.is_zero_block(block):
                continue
            spaces = split_spaces(block)
            slice_list = tuple(slices[sp] for sp in spaces)
            dm_block = self[block].to_ndarray()
            ret[slice_list] = dm_block
        return ret

    def _construct_empty(self):
        """
        Create an empty instance of an NParticleOperator
        """
        return NParticleOperator(
            self.mospaces,
            self.n_particle_op,
            symmetry=self.symmetry,
        )

    def copy(self):
        """
        Return a deep copy of the NParticleOperator
        """
        ret = self._construct_empty()
        for b in self.blocks_nonzero:
            ret[b] = self.block(b).copy()
        if self.reference_state is not None:
            ret.reference_state = self.reference_state
        return ret

    def _transform_to_ao(self, refstate):
        raise NotImplementedError("Needs to be implemented on the "
                                  f"{self.__class__.__name__} class.")

    def to_ao_basis(self, refstate=None):
        """
        Transforms the NParticleOperator to the atomic orbital
        basis using a ReferenceState. If no ReferenceState is given,
        the ReferenceState used to construct the NParticleOperator is taken instead.
        """
        if isinstance(refstate, libadcc.ReferenceState):
            return self._transform_to_ao(refstate)
        elif refstate is None:
            if self.reference_state is None:
                raise ValueError("Argument reference_state is required if no "
                                 "reference_state is stored in the "
                                 "NParticleOperator")
            return self._transform_to_ao(self.reference_state)
        else:
            raise TypeError("Argument type not supported.")

    def __iadd__(self, other):
        from . import block as b
        if self.mospaces != other.mospaces:
            raise ValueError(f"Cannot add {self.__class__.__name__}s with "
                             "differing mospaces.")
        if self.symmetry is not OperatorSymmetry.NOSYMMETRY and \
                self.symmetry is not other.symmetry:
            raise ValueError("Cannot add non-symmetric matrix "
                             "in-place to symmetric one.")

        # other.canonical_blocks is subset of self.canonical_blocks
        assert all(b in self.canonical_blocks for b in other.canonical_blocks)
        for block in other.blocks_nonzero:
            if self.is_zero_block(block):
                self[block] = other.block(block).copy()
            else:
                self[block] += other.block(block)

        if self.symmetry is OperatorSymmetry.NOSYMMETRY \
                and other.symmetry is not OperatorSymmetry.NOSYMMETRY:
            for block in self.blocks_nonzero:
                bra = block[:2 * self.n_particle_op]
                ket = block[2 * self.n_particle_op:]
                c_block, factor, transpose = b.get_canonical_block(
                    bra, ket, other.symmetry
                )
                if block == c_block:
                    continue  # Done already
                obT = 0
                if not other.is_zero_block(c_block):
                    obT = factor * other.block(c_block).transpose(transpose)
                if not self.is_zero_block(block):
                    obT += self.block(block)
                self[block] = evaluate(obT)

        # Update ReferenceState pointer
        if self.reference_state is not None:
            if other.reference_state is not None \
                    and self.reference_state != other.reference_state:
                self.reference_state = None
        return self

    def __isub__(self, other):
        from . import block as b
        if self.mospaces != other.mospaces:
            raise ValueError(f"Cannot add {self.__class__.__name__}s with "
                             "differing mospaces.")
        if self.symmetry is not OperatorSymmetry.NOSYMMETRY and \
                self.symmetry is not other.symmetry:
            raise ValueError("Cannot add non-symmetric matrix "
                             "in-place to symmetric one.")

        # other.canonical_blocks is subset of self.canonical_blocks
        assert all(b in self.canonical_blocks for b in other.canonical_blocks)
        for block in other.blocks_nonzero:
            if self.is_zero_block(block):
                self[block] = -1.0 * other.block(block)  # The copy is implicit
            else:
                self[block] -= other.block(block)

        if self.symmetry is OperatorSymmetry.NOSYMMETRY \
                and other.symmetry is not OperatorSymmetry.NOSYMMETRY:
            for block in self.blocks_nonzero:
                bra = block[:2 * self.n_particle_op]
                ket = block[2 * self.n_particle_op:]
                c_block, factor, transpose = b.get_canonical_block(
                    bra, ket, other.symmetry
                )
                if block == c_block:
                    continue  # Done already
                obT = 0
                if not other.is_zero_block(c_block):
                    obT = -1 * factor * other.block(c_block).transpose(transpose)
                if not self.is_zero_block(block):
                    obT += self.block(block)
                self[block] = evaluate(obT)

        # Update ReferenceState pointer
        if self.reference_state is not None:
            if other.reference_state is not None \
                    and self.reference_state != other.reference_state:
                self.reference_state = None
        return self

    def __imul__(self, other):
        if not isinstance(other, (float, int)):
            return NotImplemented
        for b in self.blocks_nonzero:
            self[b] = self.block(b) * other
        return self

    def __add__(self, other):
        if (
            self.symmetry is not other.symmetry
            and self.symmetry is not OperatorSymmetry.NOSYMMETRY
            and other.symmetry is not OperatorSymmetry.NOSYMMETRY
        ):
            raise ValueError("Addition of Hermitian and Antihermitian "
                             "operators is not implemented.")

        if self.symmetry is OperatorSymmetry.NOSYMMETRY \
                or other.symmetry is not OperatorSymmetry.NOSYMMETRY:
            return self.copy().__iadd__(other)
        else:
            return other.copy().__iadd__(self)

    def __sub__(self, other):
        if (
            self.symmetry is not other.symmetry
            and self.symmetry is not OperatorSymmetry.NOSYMMETRY
            and other.symmetry is not OperatorSymmetry.NOSYMMETRY
        ):
            raise ValueError("Substraction of Hermitian and Antihermitian "
                             "operators is not implemented.")

        if self.symmetry is OperatorSymmetry.NOSYMMETRY \
                or other.symmetry is not OperatorSymmetry.NOSYMMETRY:
            return self.copy().__isub__(other)
        else:
            return (-1.0 * other).__iadd__(self)

    def __mul__(self, other):
        return self.copy().__imul__(other)

    def __rmul__(self, other):
        return self.copy().__imul__(other)

    def evaluate(self):
        for b in self.blocks_nonzero:
            self.block(b).evaluate()
        return self

    def set_random(self):
        for b in self.canonical_blocks:
            self[b].set_random()
        return self


def product_trace(op1, op2) -> float:
    """
    Compute the expectation value (inner product) of two N-particle operators.

    For the special case of a density operator and a general operator, the
    inner product is identical in the AO and MO basis.

    Parameters
    ----------
    op1, op2 : NParticleOperator or NParticleDensity
        Operators or densities to compute the inner product of.

    Returns
    -------
    float
        The trace / expectation value <op1, op2>.
    """
    # TODO use blocks_nonzero and build the set intersection
    #      to avoid the is_zero_block( ) checks below.
    #      I'm a bit hesitant to do this right now, because I'm lacking
    #      the time at the moment to build a more sophisticated test,
    #      which could potentially catch an arising error.
    if op1.n_particle_op != op2.n_particle_op:
        raise TypeError(
            "Both operators must have the same number of N-particle operators "
            f"(got {op1.n_particle_op} and {op2.n_particle_op})."
        )
    assert op1.blocks == op2.blocks
    if op1.symmetry == OperatorSymmetry.NOSYMMETRY:
        factors = op1.canonical_factors.copy()
    elif op2.symmetry == OperatorSymmetry.NOSYMMETRY:
        factors = op2.canonical_factors.copy()
    else:
        assert op1.canonical_factors == op2.canonical_factors
        factors = op1.canonical_factors.copy()
        # Off-diagonal terms can be removed since a Hermitian times
        # an anti-Hermitian operator yields for off diagonal terms
        # (e.g. for a one-particle operator) ov * ov and vo * vo = (ov).T * (−ov.T),
        # which cancel exactly to zero.
        if op1.symmetry is not op2.symmetry:
            to_remove = []
            for b in list(factors.keys()):
                spaces = split_spaces(b)
                n = op1.n_particle_op
                if spaces[:n] != spaces[n:]:
                    to_remove.append(b)
            for b in to_remove:
                factors.pop(b)
    ret = 0
    for b, factor in factors.items():
        if op1.is_zero_block(b) or op2.is_zero_block(b):
            continue

        ret += (1 / math.factorial(op1.n_particle_op) ** 2 * factor
                * op1.block(b).dot(op2.block(b)))
    return ret



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2020 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
from .timings import Timer
from .functions import evaluate


class Intermediates():
    """
    Class offering to return a number of intermediate tensors.
    The tensors are cached internally for later reuse
    """
    generators = {}      # Registered generator functions

    def __init__(self, ground_state):
        self.ground_state = ground_state
        self.reference_state = ground_state.reference_state
        self.timer = Timer()
        self.cached_tensors = {}  # Cached tensors
        # TODO Make caching configurable ??

    def __getattr__(self, key):
        if key in self.cached_tensors:
            return self.cached_tensors[key]
        elif key in self.generators:
            # Evaluate the tensor, all generators take (hf, mp, intermediates)
            generator = self.generators[key]
            with self.timer.record(key):
                tensor = generator(self.reference_state, self.ground_state, self)
                self.cached_tensors[key] = evaluate(tensor)
            return self.cached_tensors[key]
        else:
            raise AttributeError

    def clear(self):
        """Clear all cached tensors to free storage"""
        self.cached_tensors.clear()

    def __repr__(self):
        return (
            "AdcIntermediates(contains="
            + list(self.cached_tensors.keys()).join(",")
            + ")"
        )


def register_as_intermediate(function):
    """
    This decorator allows to register a function such that it can
    be used to produce intermediates for storage in Intermediates.
    The rule of thumb is that this should only be used on expressions
    which are used in multiple places of the ADC code (e.g. properties
    and matrix, multiple ADC variants etc.)
    """
    Intermediates.generators[function.__name__] = function
    return function



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
import adcc

import molsturm

# Run SCF in molsturm
atoms = ["O", "H", "H"]
coords = [[0, 0, 0],
          [0, 0, 1.795239827225189],
          [1.693194615993441, 0, -0.599043184453037]]
system = molsturm.System(atoms, coords)

hfres = molsturm.hartree_fock(system, basis_type="gaussian",
                              basis_set_name="sto-3g",
                              conv_tol=1e-12, print_iterations=True)

# Run an adc2 calculation:
singlets = adcc.adc2(hfres, n_singlets=5, conv_tol=1e-9)
triplets = adcc.adc2(singlets.matrix, n_triplets=3, conv_tol=1e-9)

print(singlets.describe())
print(triplets.describe())



from __future__ import annotations
import numpy
import pybind11_stubgen.typing_ext
import typing

__all__: list[str] = [
    "AdcMemory",
    "HartreeFockProvider",
    "HartreeFockSolution_i",
    "MoIndexTranslation",
    "MoSpaces",
    "ReferenceState",
    "Symmetry",
    "Tensor",
    "amplitude_vector_enforce_spin_kind",
    "direct_sum",
    "evaluate",
    "fill_pp_doubles_guesses",
    "get_n_threads",
    "get_n_threads_total",
    "linear_combination_strict",
    "make_symmetry_eri",
    "make_symmetry_operator",
    "make_symmetry_operator_basis",
    "make_symmetry_orbital_coefficients",
    "make_symmetry_orbital_energies",
    "make_symmetry_triples",
    "set_n_threads",
    "set_n_threads_total",
    "tensordot",
    "trace",
]

class AdcMemory:
    """
    Class controlling the memory allocations for adcc ADC calculations. Python binding to :cpp:class:`libadcc::AdcMemory`.
    """
    def __init__(self) -> None: ...
    def __repr__(self) -> str: ...
    def initialise(
        self, pagefile_directory: str, max_block_size: int, allocator: str
    ) -> None: ...
    @property
    def allocator(self) -> str:
        """
        Return the allocator to which the class is initialised.
        """
    @property
    def contraction_batch_size(self) -> int:
        """
        Get or set the batch size for contraction, i.e. the number of elements handled simultaneously in a tensor contraction.
        """
    @contraction_batch_size.setter
    def contraction_batch_size(self, arg1: int) -> None: ...
    @property
    def max_block_size(self) -> int:
        """
        Return the maximal block size a tenor may have along each axis.
        """
    @property
    def pagefile_directory(self) -> str:
        """
        Return the pagefile_directory value:
        Note: This value is only meaningful if allocator != "standard"
        """

class HartreeFockProvider(HartreeFockSolution_i):
    """
    Abstract class defining the interface for passing data from the host program to adcc. All functions of this class need to be overwritten explicitly from python.
    In the remaining documentation we denote with `nf` the value returned by `get_n_orbs_alpha()` and with `nb` the value returned by `get_nbas()`.
    """
    def __init__(self) -> None: ...
    def fill_eri_ffff(self, arg0: tuple, arg1: numpy.ndarray) -> None:
        """
        Fill the passed numpy array `arg1` with a part of the electron-repulsion integral tensor in the molecular orbital basis. The indexing convention is the chemist's notation, i.e. the index tuple `(i,j,k,l)` refers to the integral :math:`(ij|kl)`. The block to store is specified by the provided tuple of ranges `arg0`, which gives the range of indices to place into the buffer along each of the axis. The index counting is done in spin orbitals, so the full range in each axis is `range(0, 2 * nf)`.
        """
    def fill_eri_phys_asym_ffff(self, arg0: tuple, arg1: numpy.ndarray) -> None:
        """
        Fill the passed numpy array `arg1` with a part of the **antisymmetrised** electron-repulsion integral tensor in the molecular orbital basis. The indexing convention is the physicist's notation, i.e. the index tuple `(i,j,k,l)` refers to the integral :math:`\\langle ij||kl \\rangle`. The block to store is specified by the provided tuple of ranges `arg0`, which gives the range of indices to place into the buffer along each of the axis. The index counting is done in spin orbitals, so the full range in each axis is `range(0, 2 * nf)`.
        """
    def fill_fock_ff(self, arg0: tuple, arg1: numpy.ndarray) -> None:
        """
        Fill the passed numpy array `arg1` with a part of the Fock matrix in the molecular orbital basis. The block to store is specified by the provided tuple of ranges `arg0`, which gives the range of indices to place into the buffer along each of the axis. The index counting is done in spin orbitals, so the full range in each axis is `range(0, 2 * nf)`. The implementation should not assume that the alpha-beta and beta-alpha blocks are not accessed even though they are zero by spin symmetry.
        """
    def fill_occupation_f(self, arg0: numpy.ndarray) -> None:
        """
        Fill the passed numpy array of size `(2 * nf, )` with the occupation number for each SCF orbital.
        """
    def fill_orbcoeff_fb(self, arg0: numpy.ndarray) -> None:
        """
        Fill the passed numpy array of size `(2 * nf, nb)` with the SCF orbital coefficients, i.e. the uniform transform from the one-particle basis to the molecular orbitals.
        """
    def fill_orben_f(self, arg0: numpy.ndarray) -> None:
        """
        Fill the passed numpy array of size `(2 * nf, )` with the SCF orbital energies.
        """
    def flush_cache(self) -> None:
        """
        This function is called to signal that potential cached data could now be flushed to save memory or other resources.
        This can be used to purge e.g. intermediates for the computation of electron-repulsion integral tensor data.
        """
    def get_backend(self) -> str:
        """
        Return the identifier of the SCF backend. Must be overridden in Python subclasses.
        """
    def get_conv_tol(self) -> float:
        """
        Returns the tolerance value used for SCF convergence. Should be roughly equivalent to the l2 norm of the Pulay error.
        """
    def get_energy_scf(self) -> float:
        """
        Returns the final total SCF energy (sum of electronic and nuclear terms.
        """
    def get_n_bas(self) -> int:
        """
        Returns the number of *spatial* one-electron basis functions. This value is abbreviated by `nb` in the documentation.
        """
    def get_n_orbs_alpha(self) -> int:
        """
        Returns the number of HF *spin* orbitals of alpha spin. It is assumed the same number of beta spin orbitals are used. This value is abbreviated by `nf` in the documentation.
        """
    def get_nuclear_multipole(
        self, arg0: int, arg1: tuple
    ) -> numpy.ndarray[numpy.float64]:
        """
        Returns the nuclear multipole of the requested order. For `0` returns the total nuclear charge as an array of size 1, for `1` returns the nuclear dipole moment as an array of size 3.
        """
    def get_restricted(self) -> bool:
        """
        Return *True* for a restricted SCF calculation, *False* otherwise.
        """
    def get_spin_multiplicity(self) -> int:
        """
        Returns the spin multiplicity of the HF ground state. A value of 0* (for unknown) should be supplied for unrestricted calculations.
        """
    def has_eri_phys_asym_ffff(self) -> bool:
        """
        Returns whether `fill_eri_phys_asym_ffff` function is implemented and should be used(*True*) or whether antisymmetrisation should be done inside adcc starting from the `fill_eri_ffff` function (*False*)
        """
    def transform_gauge_origin_to_xyz(self, arg0: str) -> tuple:
        """
        Transforms a string specifying the gauge origin to a tuple containing the x, y, z Cartesian components.
        """

class HartreeFockSolution_i:
    """
    Interface class representing the data expected in adcc from an interfacing HF / SCF program. Python binding to :cpp:class:`HartreeFockSolution_i`
    """
    @property
    def backend(self) -> str: ...
    @property
    def conv_tol(self) -> float: ...
    @property
    def energy_scf(self) -> float: ...
    @property
    def fock_ff(self) -> numpy.ndarray[numpy.float64]: ...
    @property
    def n_alpha(self) -> int: ...
    @property
    def n_bas(self) -> int: ...
    @property
    def n_beta(self) -> int: ...
    @property
    def n_orbs(self) -> int: ...
    @property
    def n_orbs_alpha(self) -> int: ...
    @property
    def n_orbs_beta(self) -> int: ...
    @property
    def occupation_f(self) -> numpy.ndarray[numpy.float64]: ...
    @property
    def orbcoeff_fb(self) -> numpy.ndarray[numpy.float64]: ...
    @property
    def orben_f(self) -> numpy.ndarray[numpy.float64]: ...
    @property
    def restricted(self) -> bool: ...
    @property
    def spin_multiplicity(self) -> int: ...

class MoIndexTranslation:
    """
    Helper object to extract information from indices into orbitals subspaces and to map them between different indexing conventions (full MO space, MO subspaces, indexing convention in the HF Provider / SCF host program, ... Python binding to :cpp:class:`libadcc::MoIndexTranslation`.
    """
    @typing.overload
    def __init__(self, arg0: MoSpaces, arg1: str) -> None:
        """
        Construct a MoIndexTranslation class from an MoSpaces object and the identifier for the space (e.g. o1o1, v1o1, o3v2o1v1, ...)
        """
    @typing.overload
    def __init__(self, arg0: MoSpaces, arg1: list[str]) -> None:
        """
        Construct a MoIndexTranslation class from an MoSpaces object and the list of identifiers for the space (e.g. ["o1", "o1"] ...)
        """
    def block_index_of(self, arg0: tuple) -> tuple:
        """
        Get the block index of an index, i.e. get the index which points to the block of the tensor in which the element with the passed index is contained in.
        """
    def block_index_spatial_of(self, arg0: tuple) -> tuple:
        """
        Get the spatial block index of an index

        The spatial block index is the result of block_index_of modulo the spin blocks,
        i.e. it maps an index onto the index of the *spatial* blocks only, such that
        the resulting value is identical for two index where the MOs only differ
        by spin. For example the 1st core alpha and the 1st core beta orbital will
        map to the same value upon a call of this function.
        """
    @typing.overload
    def combine(self, arg0: tuple, arg1: tuple) -> tuple:
        """
        Combine a block index and an in-block index into the appropriate index. Effectively undoes the effect of 'split'.
        """
    @typing.overload
    def combine(self, arg0: str, arg1: tuple, arg2: tuple) -> tuple:
        """
        Combine a spin block (given as a string of 'a's or 'b's), a spatial-only block index and an in-block index into the appropriate index. Essentially undoes the effect of 'spin_of', 'block_index_spatial_of' and 'inblock_index_of'.
        """
    def full_index_of(self, arg0: tuple) -> tuple:
        """
        Map an index given in the space, which was passed upon construction, to the corresponding index in the full MO index range (the ffff space).
        """
    def hf_provider_index_of(self, arg0: tuple) -> tuple:
        """
        Map an index (given in the space passed upon construction) to the indexing convention of the host program provided to adcc as the HF provider.
        """
    def inblock_index_of(self, arg0: tuple) -> tuple:
        """
        Get the in-block index, i.e. the index within the tensor block.
        """
    def map_range_to_hf_provider(self, arg0: tuple) -> list:
        """
        Map a range of indices to host program indices, i.e. the indexing convention
        used in the HfProvider, which provides the SCF data to adcc.

        Since the mapping between subspace and host program indices might not be contiguous,
        a list of pairs of ranges is returned. In each pair, the first entry represents a
        range of indices (indexed in the MO subspace) and the second entry represents
        the equivalent range of indices in the Hartree-Fock provider these are mapped to.

          ranges    Tuple of pairs of indices: One index pair for each dimension. Each
                    pair describes the range of indices along one axis, which should be
                    mapped to the indexing convention of the HfProvider. The range should
                    be thought of as a half-open interval [start, end), where start and
                    end are the indexed passed as a pair to the function.
        """
    def spin_of(self, arg0: tuple) -> str:
        """
        Get the spin block of each of the index components as a string.
        """
    def split(self, arg0: tuple) -> tuple:
        """
        Split an index into block index and in-block index
        """
    def split_spin(self, arg0: tuple) -> tuple:
        """
        Split an index into a spin block descriptor, a spatial block index and an in-block index.
        """
    @property
    def mospaces(self) -> MoSpaces:
        """
        Return the MoSpaces object supplied on initialisation
        """
    @property
    def ndim(self) -> int:
        """
        Return the number of dimensions.
        """
    @property
    def shape(self) -> tuple:
        """
        Return the length along each dimension.
        """
    @property
    def space(self) -> str:
        """
        Return the space supplied on initialisation.
        """
    @property
    def subspaces(self) -> list[str]: ...

class MoSpaces:
    """
    Class setting up the molecular orbital index spaces and subspaces and exposing information about them. Python binding to :cpp:class:`libadcc::MoSpaces`.
    """
    def __init__(
        self,
        arg0: HartreeFockSolution_i,
        arg1: AdcMemory,
        arg2: list[int],
        arg3: list[int],
        arg4: list[int],
    ) -> None:
        """
        Construct an MoSpaces object from a HartreeFockSolution_i, a pointer to
        an AdcMemory object.

        adcmem_ptr        ADC memory keep-alive object to be used in all Tensors
                          constructed using this MoSpaces object.
        core_orbitals     List of orbitals indices (in the full fock space, original
                          ordering of the hf object), which defines the orbitals to
                          be put into the core space, if any. The same number
                          of alpha and beta orbitals should be selected. These will
                          be forcibly occupied.
        frozen_core_orbitals
                          List of orbital indices, which define the frozen core,
                          i.e. those occupied orbitals, which do not take part in
                          the ADC calculation. The same number of alpha and beta
                          orbitals has to be selected.
        frozen_virtuals    List of orbital indices, which the frozen virtuals,
                          i.e. those virtual orbitals, which do not take part
                          in the ADC calculation. The same number of alpha and beta
                          orbitals has to be selected.
        """
    def n_orbs(self, arg0: str) -> int:
        """
        The number of orbitals in a particular orbital subspace
        """
    def n_orbs_alpha(self, arg0: str) -> int:
        """
        The number of alpha orbitals in a particular orbital subspace
        """
    def n_orbs_beta(self, arg0: str) -> int:
        """
        The number of beta orbitals in a particular orbital subspace
        """
    @property
    def has_core_occupied_space(self) -> bool:
        """
        Does this object have a core-occupied space (i.e. is it ready for core-valence separation)?
        """
    @property
    def irrep_totsym(self) -> str:
        """
        Return the totally symmetric irreducible representation.
        """
    @property
    def irreps(self) -> list[str]:
        """
        The irreducible representations in the point group to which this class has been set up.
        """
    @property
    def map_block_irrep(self) -> dict[str, list[str]]:
        """
        Contains for each orbital space the mapping from each *block* used inside the space to the irreducible representation it correspond to.
        """
    @property
    def map_block_spin(self) -> dict[str, list[str]]:
        """
        Contains for each orbital space the mapping from each *block* used inside the space to the spin it correspond to ('a' is alpha and 'b' is beta)
        """
    @property
    def map_block_start(self) -> dict[str, list[int]]:
        """
        Contains for each orbital space the indices at which a new tensor block starts. Thus this list contains at least on index, namely 0.
        """
    @property
    def map_index_hf_provider(self) -> dict[str, list[int]]:
        """
        Contains for each orbital space (e.g. f, o1) a mapping from the indices used inside
        the respective space to the molecular orbital index convention used in the host
        program, i.e. to the ordering in which the molecular orbitals have been exposed
        in the HartreeFockSolution_i object passed on class construction.
        """
    @property
    def point_group(self) -> str:
        """
        The name of the point group for which the data in this class has been set up.
        """
    @property
    def restricted(self) -> bool:
        """
        Are the orbitals resulting from a restricted SCF calculation, such that alpha and beta electron share the same spatial part.
        """
    @property
    def subspaces(self) -> list[str]:
        """
        The list of all orbital subspaces known to this object.
        """
    @property
    def subspaces_occupied(self) -> list[str]:
        """
        The list of occupied orbital subspaces known to this object.
        """
    @property
    def subspaces_virtual(self) -> list[str]:
        """
        The list of virtual orbital subspaces known to this object.
        """

class ReferenceState:
    """
    Class representing information about the reference state for adcc. Python binding to:cpp:class:`libadcc::ReferenceState`.
    """
    def __init__(self, arg0: HartreeFockSolution_i, arg1: MoSpaces, arg2: bool) -> None:
        """
        Setup a ReferenceStateject using an MoSpaces object.

        hfsoln_ptr        Pointer to the Interface to the host program,
                          providing the HartreeFockSolution data, which
                          will be provided by this object.
        mo_ptr            MoSpaces object containing info about the MoSpace setup
                          and the point group symmetry.
        symmetry_check_on_import
                          Should symmetry of the imported objects be checked
                          explicitly during the import process. This massively slows
                          down the import process and has a dramatic impact on memory
                          usage and should thus only be used for debugging import routines
                          from the host programs. Do not enable this unless you know
                          that you really want to.
        """
    def eri(self, arg0: str) -> Tensor:
        """
        Return the ERI (electron-repulsion integrals) tensor block corresponding to the provided space.
        """
    def flush_hf_cache(self) -> None:
        """
        Tell the contained HartreeFockSolution_i object (which was passed upon construction), that a larger amount of import operations is done and that the next request for further imports will most likely take some time, such that intermediate caches can now be flushed to save some memory or other resources.
        """
    def fock(self, arg0: str) -> Tensor:
        """
        Return the Fock matrix block corresponding to the provided space.
        """
    def gauge_origin_to_xyz(self, arg0: str) -> tuple: ...
    def import_all(self) -> None:
        """
        Normally the class only imports the Fock matrix blocks and electron-repulsion integrals of a particular space combination when this is requested by a call to above fock() or eri() functions. This function call, however, instructs the class to immediately import *all* such blocks. Typically you do not want to do this.
        """
    def nuclear_quadrupole(
        self,
        arg0: typing.Annotated[list[float], pybind11_stubgen.typing_ext.FixedSize(3)],
    ) -> numpy.ndarray[numpy.float64]: ...
    def orbital_coefficients(self, arg0: str) -> Tensor:
        """
        Return the molecular orbital coefficients corresponding to the provided space (alpha and beta coefficients are returned)
        """
    def orbital_coefficients_alpha(self, arg0: str) -> Tensor:
        """
        Return the alpha molecular orbital coefficients corresponding to the provided space
        """
    def orbital_coefficients_beta(self, arg0: str) -> Tensor:
        """
        Return the beta molecular orbital coefficients corresponding to the provided space
        """
    def orbital_energies(self, arg0: str) -> Tensor:
        """
        Return the orbital energies corresponding to the provided space
        """
    @property
    def backend(self) -> str:
        """
        The identifier of the back end used for the SCF calculation.
        """
    @property
    def cached_eri_blocks(self) -> list[str]:
        """
        Get or set the list of momentarily cached ERI tensor blocks

        Setting this property allows to drop ERI tensor blocks if they are no longer needed to save memory.
        """
    @cached_eri_blocks.setter
    def cached_eri_blocks(self, arg1: list[str]) -> None: ...
    @property
    def cached_fock_blocks(self) -> list[str]:
        """
        Get or set the list of momentarily cached Fock matrix blocks

        Setting this property allows to drop fock matrix blocks if they are no longer needed to save memory.
        """
    @cached_fock_blocks.setter
    def cached_fock_blocks(self, arg1: list[str]) -> None: ...
    @property
    def conv_tol(self) -> float:
        """
        SCF convergence tolererance
        """
    @property
    def energy_scf(self) -> float:
        """
        Final total SCF energy
        """
    @property
    def has_core_occupied_space(self) -> bool:
        """
        Is a core occupied space setup, such that a core-valence separation can be applied.
        """
    @property
    def irreducible_representation(self) -> str:
        """
        Reference state irreducible representation
        """
    @property
    def mospaces(self) -> MoSpaces:
        """
        The MoSpaces object supplied on initialisation
        """
    @property
    def n_alpha(self) -> int:
        """
        Number of alpha electrons
        """
    @property
    def n_beta(self) -> int:
        """
        Number of beta electrons
        """
    @property
    def n_orbs(self) -> int:
        """
        Number of molecular orbitals
        """
    @property
    def n_orbs_alpha(self) -> int:
        """
        Number of alpha orbitals
        """
    @property
    def n_orbs_beta(self) -> int:
        """
        Number of beta orbitals
        """
    @property
    def nuclear_dipole(self) -> numpy.ndarray[numpy.float64]: ...
    @property
    def nuclear_total_charge(self) -> float: ...
    @property
    def restricted(self) -> bool:
        """
        Return whether the reference is restricted or not.
        """
    @property
    def spin_multiplicity(self) -> int:
        """
        Return the spin multiplicity of the reference state. 0 indicates that the spin cannot be determined or is not integer (e.g. UHF)
        """
    @property
    def timer(self) -> typing.Any:
        """
        Obtain the timer object of this class.
        """

class Symmetry:
    """
    Container for Tensor symmetry information
    """
    @typing.overload
    def __init__(self, arg0: MoSpaces, arg1: str) -> None:
        """
        Construct a Symmetry class from an MoSpaces object and the identifier for the space (e.g. o1o1, v1o1, o3v2o1v1, ...). Python binding to :cpp:class:`libadcc::Symmetry`.
        """
    @typing.overload
    def __init__(
        self, arg0: MoSpaces, arg1: str, arg2: dict[str, tuple[int, int]]
    ) -> None:
        """
        Construct a Symmetry class from an MoSpaces object, a space string and a map to supply the number of orbitals for some additional axes.
        For the additional axis the pair contains either two numbers (for the number of alpha and beta orbitals in this axis) or only one number and a zero (for an axis, which as only one spin kind, alpha or beta).

        This is an advanced constructor. Use only if you know what you do.
        """
    def clear(self) -> None:
        """
        Clear the symmetry.
        """
    def describe(self) -> str:
        """
        Return a descriptive string.
        """
    @property
    def empty(self) -> bool:
        """
        Is the symmetry empty (i.e. noy symmetry setup)
        """
    @property
    def irreps_allowed(self) -> list[str]:
        """
        The list of irreducible representations, for which the tensor shall be non-zero. If this is *not* set, i.e. an empty list, all irreps will be allowed.
        """
    @irreps_allowed.setter
    def irreps_allowed(self, arg1: list[str]) -> None: ...
    @property
    def mospaces(self) -> MoSpaces:
        """
        Return the MoSpaces object supplied on initialisation
        """
    @property
    def ndim(self) -> int:
        """
        Return the number of dimensions.
        """
    @property
    def permutations(self) -> list[str]:
        """
        The list of index permutations, which do not change the tensor.
        A minus may be used to indicate anti-symmetric
        permutations with respect to the first (reference) permutation.

        For example the list ["ij", "ji"] defines a symmetric matrix
        and ["ijkl", "-jikl", "-ijlk", "klij"] the symmetry of the ERI
        tensor. Not all permutations need to be given to fully describe
        the symmetry. Beware that the check for errors and conflicts
        is only rudimentary at the moment.
        """
    @permutations.setter
    def permutations(self, arg1: list[str]) -> None: ...
    @property
    def shape(self) -> tuple:
        """
        Return the shape of tensors constructed from this symmetry.
        """
    @property
    def space(self) -> str:
        """
        Return the space supplied on initialisation.
        """
    @property
    def spin_block_maps(self) -> list[tuple[str, str, float]]:
        """
        A list of tuples of the form ("aaaa", "bbbb", -1.0), i.e.
        two spin blocks followed by a factor. This maps the second onto the first
        with a factor of -1.0 between them.
        """
    @spin_block_maps.setter
    def spin_block_maps(self, arg1: list[tuple[str, str, float]]) -> None: ...
    @property
    def spin_blocks_forbidden(self) -> list[str]:
        """
        List of spin-blocks, which are marked forbidden (i.e. enforce them to stay zero).
        Blocks are given as a string in the letters 'a' and 'b', e.g. ["aaba", "abba"]
        """
    @spin_blocks_forbidden.setter
    def spin_blocks_forbidden(self, arg1: list[str]) -> None: ...

class Tensor:
    """
    Class representing the Tensor objects used for computations in adcc
    """

    flags: list[str]
    @typing.overload
    def __add__(self, arg0: float) -> Tensor: ...
    @typing.overload
    def __add__(self, arg0: Tensor) -> Tensor: ...
    def __getitem__(self, arg0: tuple) -> float:
        """
        Get a tensor element or a slice of tensor elements.
        """
    def __iadd__(self, arg0: Tensor) -> Tensor: ...
    def __imul__(self, arg0: float) -> Tensor: ...
    def __init__(self, arg0: Symmetry) -> None:
        """
        Construct a Tensor object using a Symmetry object describing its symmetry properties.
        The returned object is not guaranteed to contain initialised memory. Python binding to :cpp:class:`libadcc::Tensor`
        """
    def __isub__(self, arg0: Tensor) -> Tensor: ...
    def __itruediv__(self, arg0: float) -> Tensor: ...
    def __len__(self) -> int: ...
    def __matmul__(self, arg0: Tensor) -> Tensor: ...
    @typing.overload
    def __mul__(self, arg0: float) -> Tensor: ...
    @typing.overload
    def __mul__(self, arg0: Tensor) -> Tensor:
        """
        Multiply two tensors elementwise.
        """
    def __neg__(self) -> Tensor: ...
    def __pos__(self) -> Tensor: ...
    def __radd__(self, arg0: float) -> Tensor: ...
    def __repr__(self) -> typing.Any: ...
    def __rmul__(self, arg0: float) -> Tensor: ...
    def __rsub__(self, arg0: float) -> Tensor: ...
    def __setitem__(self, arg0: tuple, arg1: float) -> float:
        """
        Set a tensor element or a slice of tensor elements. The operation will adhere symmetry, i.e. alter all elements equivalent by symmetry at once.
        """
    def __str__(self) -> typing.Any: ...
    @typing.overload
    def __sub__(self, arg0: float) -> Tensor: ...
    @typing.overload
    def __sub__(self, arg0: Tensor) -> Tensor: ...
    @typing.overload
    def __truediv__(self, arg0: float) -> Tensor: ...
    @typing.overload
    def __truediv__(self, arg0: Tensor) -> Tensor:
        """
        Divide two tensors elementwise.
        """
    @typing.overload
    def antisymmetrise(self, arg0: list) -> Tensor: ...
    @typing.overload
    def antisymmetrise(self, *args) -> Tensor: ...
    def copy(self) -> Tensor:
        """
        Returns a deep copy of the tensor.
        """
    @typing.overload
    def describe_expression(self, arg0: str) -> str:
        """
        Return a string providing a hopefully descriptive representation of the tensor expression stored inside the object.
        """
    @typing.overload
    def describe_expression(self) -> str: ...
    def describe_symmetry(self) -> str:
        """
        Return a string providing a hopefully descriptive representation of the symmetry information stored inside the tensor.
        """
    def diagonal(self, *args) -> Tensor: ...
    @typing.overload
    def dot(self, arg0: Tensor) -> float: ...
    @typing.overload
    def dot(self, arg0: list) -> numpy.ndarray[numpy.float64]: ...
    def empty_like(self) -> Tensor: ...
    def evaluate(self) -> Tensor:
        """
        Ensure the tensor to be fully evaluated and resilient in memory. Usually happens automatically when needed. Might be useful for fine-tuning, however.
        """
    def is_allowed(self, arg0: tuple) -> bool:
        """
        Is a particular index allowed by symmetry
        """
    def nosym_like(self) -> Tensor: ...
    def ones_like(self) -> Tensor: ...
    def select_n_absmax(self, arg0: int) -> list:
        """
        Select the n absolute maximal elements.
        """
    def select_n_absmin(self, arg0: int) -> list:
        """
        Select the n absolute minimal elements.
        """
    def select_n_max(self, arg0: int) -> list:
        """
        Select the n maximal elements.
        """
    def select_n_min(self, arg0: int) -> list:
        """
        Select the n minimal elements.
        """
    @typing.overload
    def set_from_ndarray(self, arg0: numpy.ndarray) -> Tensor:
        """
        Set all tensor elements from a standard np::ndarray by making a copy. Provide an optional tolerance argument to increase the tolerance for the check for symmetry consistency.
        """
    @typing.overload
    def set_from_ndarray(
        self, arg0: numpy.ndarray[numpy.float64], arg1: float
    ) -> Tensor:
        """
        Set all tensor elements from a standard np::ndarray by making a copy. Provide an optional tolerance argument to increase the tolerance for the check for symmetry consistency.
        """
    def set_immutable(self) -> None:
        """
        Set the tensor as immutable, allowing some optimisations to be performed.
        """
    def set_mask(self, arg0: str, arg1: float) -> None:
        """
        Set all elements corresponding to an index mask, which is given by a string eg. 'iijkli' sets elements T_{iijkli}
        """
    def set_random(self) -> Tensor:
        """
        Set all tensor elements to random data, adhering to the internal symmetry.
        """
    @typing.overload
    def symmetrise(self, arg0: list) -> Tensor: ...
    @typing.overload
    def symmetrise(self, *args) -> Tensor: ...
    def to_ndarray(self) -> numpy.ndarray[numpy.float64]:
        """
        Export the tensor data to a standard np::ndarray by making a copy.
        """
    @typing.overload
    def transpose(self) -> Tensor: ...
    @typing.overload
    def transpose(self, arg0: tuple) -> Tensor: ...
    def zeros_like(self) -> Tensor: ...
    @property
    def T(self) -> Tensor: ...
    @property
    def mutable(self) -> bool: ...
    @property
    def ndim(self) -> int: ...
    @property
    def needs_evaluation(self) -> bool:
        """
        Does the tensor need evaluation or is it fully evaluated and resilient in memory.
        """
    @property
    def shape(self) -> tuple: ...
    @property
    def size(self) -> int: ...
    @property
    def space(self) -> str: ...
    @property
    def subspaces(self) -> list[str]: ...

def amplitude_vector_enforce_spin_kind(arg0: Tensor, arg1: str, arg2: str) -> None:
    """
    Apply the spin symmetrisation required to make the doubles and higher parts of an amplitude vector consist of components for a particular spin kind only.
    """

def direct_sum(a: Tensor, b: Tensor) -> Tensor: ...
def evaluate(arg0: Tensor) -> Tensor: ...
def fill_pp_doubles_guesses(
    guesses_d: list[Tensor],
    mospaces: MoSpaces,
    df02: Tensor,
    df13: Tensor,
    spin_change_twice: int,
    degeneracy_tolerance: float,
) -> int:
    """
    Fill the passed vector of doubles blocks with doubles guesses using the delta-Fock matrices df02 and df13, which are the two delta-Fock matrices involved in the doubles block.

    guesses_d    Vectors of guesses, all elements are assumed to be initialised to zero and the symmetry is assumed to be properly set up.
    mospaces     Mospaces object
    df02         Delta-Fock between spaces 0 and 2 of the ADC matrix
    df13         Delta-Fock between spaces 1 and 3 of the ADC matrix
    spin_change_twice   Twice the value of the spin change to enforce in an excitation.
    degeneracy_tolerance  Tolerance for two entries of the diagonal to be considered degenerate, i.e. identical.
    Returns     The number of guess vectors which have been properly initialised (the others are invalid and should be discarded).
    """

def get_n_threads() -> int:
    """
    Get the number of running worker threads used by adcc.
    """

def get_n_threads_total() -> int:
    """
    Get the total number of threads (running and sleeping) used by adcc. This will disappear in the future. Do not rely on it.
    """

def linear_combination_strict(
    coefficients: numpy.ndarray[numpy.float64], tensors: list
) -> Tensor: ...
def make_symmetry_eri(arg0: MoSpaces, arg1: str) -> Symmetry:
    """
    Return the Symmetry object like it would be set up for the passed subspace
    of the electron-repulsion tensor.

      mospaces    MoSpaces object
      space       Space string (e.g. o1v1o1v1)
    """

def make_symmetry_operator(arg0: MoSpaces, arg1: str, arg2: str, arg3: str) -> Symmetry:
    """
    Return the Symmetry object for an orbital subspace block of a one-particle operator

      mospaces    MoSpaces object
      space       Space string (e.g. o1v1)
      symmetry    Describes the symmetry of the tensor (only in effect if both
                  subspaces of the space string are identical).
                  Valid are "nosymmetry", "hermitian" and "antihermitian".
      cartesian_transformation
                  The cartesian function according to which the operator transforms.

    Valid cartesian_transformation values include:
         "1"                   Totally symmetric (default)
         "x", "y", "z"         Coordinate axis
         "xx", "xy", "yz" ...  Products of two coordinate axis
         "Rx", "Ry", "Rz"      Rotations about the coordinate axis
    """

def make_symmetry_operator_basis(
    arg0: MoSpaces, arg1: int, arg2: str, arg3: int, arg4: str
) -> Symmetry:
    """
    Return the symmetry object for an operator in the AO basis. The object will
    represent a block-diagonal matrix of the form
        ( M 0 )
        ( 0 M ).
    where M is an n_bas x n_bas block and is indentical in upper-left
    and lower-right.

    mospaces_ptr      MoSpaces pointer
    n_bas             Number of AO basis functions
    operator_symmetry Is the tensor symmetric (hermitian/antihermitian, only
                      in effect if both space axes identical).
                      Nosymmetry disables a setup of permutational symmetry.
    n_particle_op     NParticleOperator
    blocks            Which blocks of the operator to return. Valid values
                      are 'ab' to return a tensor for both alpha and beta
                      block as a block-diagonal tensor, 'a' to only return a
                      tensor for only alpha block.
    """

def make_symmetry_orbital_coefficients(
    arg0: MoSpaces, arg1: str, arg2: int, arg3: str
) -> Symmetry:
    """
    Return the Symmetry object like it would be set up for the passed subspace
    of the orbital coefficients tensor.

      mospaces    MoSpaces object
      space       Space string (e.g. o1b)
      n_bas       Number of basis functions
      blocks      Spin blocks to include. Valid are "ab", "a" and "b".
    """

def make_symmetry_orbital_energies(arg0: MoSpaces, arg1: str) -> Symmetry:
    """
    Return the Symmetry object like it would be set up for the passed subspace
    of the orbital energies tensor.

      mospaces    MoSpaces object
      space       space string (e.g. o1)
    """

def make_symmetry_triples(arg0: MoSpaces, arg1: str) -> Symmetry:
    """
    Return the Symmetry object like it would be set up for the passed subspace
    of a triples amplitude tensor.

      mospaces    MoSpaces object
      space       Space string (e.g. o1o1o1v1v1v1)
    """

def set_n_threads(arg0: int) -> None:
    """
    Set the number of running worker threads used by adcc
    """

def set_n_threads_total(arg0: int) -> None:
    """
    Set the total number of threads (running and sleeping) used by adcc. This will disappear in the future. Do not rely on it.
    """

@typing.overload
def tensordot(a: Tensor, b: Tensor, axes: typing.Iterable) -> typing.Any: ...
@typing.overload
def tensordot(a: Tensor, b: Tensor, axes: int) -> typing.Any: ...
@typing.overload
def tensordot(a: Tensor, b: Tensor) -> typing.Any: ...
@typing.overload
def trace(subscripts: str, tensor: Tensor) -> float: ...
@typing.overload
def trace(tensor: Tensor) -> float: ...



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2019 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
import numpy as np
import textwrap
from itertools import product

from .backends import import_scf_results
from .memory_pool import memory_pool

from collections.abc import Iterable

import libadcc

__all__ = ["MoSpaces"]

__valid_spaces = list(map(''.join, product(["o", "v"], ["1", "2", "3"])))


def split_spaces(spacestr):
    """Utility function to split a string with
    multiple spaces into the individual space strings
    """
    sp = textwrap.wrap(spacestr, 2)
    try:
        assert all(s in __valid_spaces for s in sp)
        return sp
    except AssertionError:
        raise ValueError(f"Invalid space string '{spacestr}'.")


def expand_spaceargs(hfdata_or_n_orbs, **spaceargs):
    if isinstance(hfdata_or_n_orbs, tuple):
        (noa, nob) = hfdata_or_n_orbs
    else:
        hfdata = hfdata_or_n_orbs
        noa = hfdata.n_orbs_alpha
        nob = hfdata.n_orbs_beta

    if isinstance(spaceargs.get("frozen_core", None), bool) \
       and spaceargs.get("frozen_core", None):
        # Determine number of frozen core electrons automatically
        # TODO The idea is to look at the energy gap in the HF orbital
        #      energies and exclude the ones, which are very far from the
        #      HOMO-LUMO gap.
        raise NotImplementedError("Automatic determination of frozen-core "
                                  "electrons not implemented.")
        #
        # TODO One could also adopt the idea in the paper by Andreas and Chong
        #      how to automatically select the frozen_virtual orbitals in a
        #      clever way.
        #

    def expand_to_list(space, entry, from_min=None, from_max=None):
        if from_min is None and from_max is None:
            raise ValueError("Both from_min and from_max is None")
        if entry is None:
            return np.array([])
        elif isinstance(entry, int):
            if from_min is not None:
                return from_min + np.arange(entry)
            return np.arange(from_max - entry, from_max)
        elif isinstance(entry, Iterable):
            return np.array(entry)
        else:
            raise TypeError("Unsupported type {} passed to argument {}"
                            "".format(type(entry), space))

    for key in spaceargs:
        if not isinstance(spaceargs[key], tuple):
            spaceargs[key] = (spaceargs[key], spaceargs[key])

    any_iterable = False
    for key in spaceargs:
        for spin in [0, 1]:
            if isinstance(spaceargs[key][spin], Iterable):
                any_iterable = True
            elif spaceargs[key][spin] is not None:
                if any_iterable:
                    raise ValueError("If one of the values of frozen_core, "
                                     "core_orbitals, frozen_virtual is an "
                                     "iterable, all must be.")

    n_orbs = [0, 0]
    for key in ["frozen_core", "core_orbitals"]:
        if key not in spaceargs or not spaceargs[key]:
            continue
        list_alpha = expand_to_list(key, spaceargs[key][0],
                                    from_min=n_orbs[0])
        list_beta = noa + expand_to_list(key, spaceargs[key][1],
                                         from_min=n_orbs[1])
        spaceargs[key] = np.concatenate((list_alpha, list_beta)).tolist()
        n_orbs[0] += len(list_alpha)
        n_orbs[1] += len(list_beta)

    key = "frozen_virtual"
    if key in spaceargs and spaceargs[key]:
        spaceargs[key] = np.concatenate((
            expand_to_list(key, spaceargs[key][0], from_max=noa),
            expand_to_list(key, spaceargs[key][1], from_max=nob) + noa
        )).tolist()
    return spaceargs


class MoSpaces(libadcc.MoSpaces):
    def __init__(self, hfdata, core_orbitals=None, frozen_core=None,
                 frozen_virtual=None):
        """Construct an MoSpaces object, which holds information for
        translating between the adcc convention of arranging molecular
        orbitals and the convention of the host program.

        The documentation of each field here is only brief. Details
        can be found in :py:`adcc.ReferenceState.__init__`.

        Parameters
        ----------
        hfdata
            Host-program object with Hartree-Fock data.

        core_orbitals : int or list or tuple, optional
            The orbitals to be put into the core-occupied space.

        frozen_core : int or list or tuple, optional
            The orbitals to be put into the frozen core space.

        frozen_virtuals : int or list or tuple, optional
            The orbitals to be put into the frozen virtual space.
        """
        if not isinstance(hfdata, libadcc.HartreeFockSolution_i):
            hfdata = import_scf_results(hfdata)

        spaceargs = expand_spaceargs(hfdata, frozen_core=frozen_core,
                                     frozen_virtual=frozen_virtual,
                                     core_orbitals=core_orbitals)
        super().__init__(hfdata, memory_pool, spaceargs["core_orbitals"],
                         spaceargs["frozen_core"], spaceargs["frozen_virtual"])

    @property
    def frozen_core(self):
        """
        The frozen (inactive) occupied spin orbitals
        (in the index convention of the host provider).
        """
        return self.map_index_hf_provider.get("o3", [])

    @property
    def core_orbitals(self):
        """
        The spin orbitals selected to reside in the core space
        (in the index convention of the host provider).
        """
        return self.map_index_hf_provider.get("o2", [])

    @property
    def occupied_orbitals(self):
        """
        The active valence-occupied spin orbitals
        (in the index convention of the host provider).
        """
        return self.map_index_hf_provider["o1"]

    @property
    def virtual_orbitals(self):
        """
        The active virtual spin orbitals
        (in the index convention of the host provider).
        """
        return self.map_index_hf_provider["v1"]

    @property
    def frozen_virtual(self):
        """
        The frozen (inactive) virtual spin orbitals
        (in the index convention of the host provider).
        """
        return self.map_index_hf_provider.get("v2", [])


# TODO some nice describe method
#      (should return the core_orbitals, frozen_core, ...)



#!/usr/bin/env python3
## vi: tabstop=4 shiftwidth=4 softtabstop=4 expandtab
## ---------------------------------------------------------------------
##
## Copyright (C) 2020 by the adcc authors
##
## This file is part of adcc.
##
## adcc is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published
## by the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## adcc is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with adcc. If not, see <http://www.gnu.org/licenses/>.
##
## ---------------------------------------------------------------------
import libadcc
import numpy as np

from adcc import evaluate
from libadcc import MoIndexTranslation
from itertools import groupby

from ..AdcMatrix import AdcMatrixlike
from .guess_zero import guess_zero


def guesses_from_diagonal(matrix, n_guesses, block="ph", spin_change=0,
                          spin_block_symmetrisation="none",
                          degeneracy_tolerance=1e-14, max_diagonal_value=1000):
    """
    Obtain guesses by inspecting a block of the diagonal of the passed ADC
    matrix. The symmetry of the returned vectors is already set-up properly.
    Note that this routine may return fewer vectors than requested in case the
    requested number could not be found.

    matrix       The matrix for which guesses are to be constructed
    n_guesses    The number of guesses to be searched for. Less number of
                 vectors are returned if this many could not be found.
    block        Diagonal block to use for obtaining the guesses
                 (typically "ph" or "pphh").
    spin_change  The spin change to enforce in an excitation.
                 Typical values are 0 (singlet/triplet/any) and -1 (spin-flip).
    spin_block_symmetrisation
                 Symmetrisation to enforce between equivalent spin blocks, which
                 all yield the desired spin_change. E.g. if spin_change == 0,
                 then both the alpha->alpha and beta->beta blocks of the singles
                 part of the excitation vector achieve a spin change of 0.
                 The symmetry specified with this parameter will then be imposed
                 between the a-a and b-b blocks. Valid values are "none",
                 "symmetric" and "antisymmetric", where "none" enforces
                 no particular symmetry.
    degeneracy_tolerance
                 Tolerance for two entries of the diagonal to be considered
                 degenerate, i.e. identical.
    max_diagonal_value
                 Maximal diagonal value, which is considered as a valid candidate
                 to form a guess.
    """
    if not isinstance(matrix, AdcMatrixlike):
        raise TypeError("matrix needs to be of type AdcMatrixlike")
    if spin_block_symmetrisation not in ["none", "symmetric", "antisymmetric"]:
        raise ValueError("Invalid value for spin_block_symmetrisation: "
                         "{}".format(spin_block_symmetrisation))
    if spin_block_symmetrisation != "none" and \
       not matrix.reference_state.restricted:
        raise ValueError("spin_block_symmetrisation != none is only valid for "
                         "ADC calculations on top of restricted reference "
                         "states.")
    if int(spin_change * 2) / 2 != spin_change:
        raise ValueError("Only integer or half-integer spin_change is allowed. "
                         "You passed {}".format(spin_change))

    if block not in matrix.axis_blocks:
        raise ValueError("The passed ADC matrix does not have the block '{}.'"
                         "".format(block))
    if n_guesses == 0:
        return []

    if block == "ph":
        guessfunction = guesses_from_diagonal_singles
    elif block == "pphh":
        guessfunction = guesses_from_diagonal_doubles
    else:
        raise ValueError(f"Don't know how to generate guesses for block {block}")

    return guessfunction(matrix, n_guesses, spin_change,
                         spin_block_symmetrisation, degeneracy_tolerance,
                         max_diagonal_value)


class TensorElement:
    def __init__(self, motrans, index, value):
        """
        Initialise a TensorElement from an MoIndexTranslation object,
        a tensor index and the corresponding value.
        """
        # TODO One could probably rewrite motrans to only take the data from
        #      the axis info the tensors have anyway. This way one could
        #      initialise such a TensorElement just from a tensor and an
        #      index, which seems like a lot more reasonable interface.
        self.index = tuple(index)
        self.subspaces = motrans.subspaces
        self.value = value

        splitted = motrans.split_spin(tuple(index))
        self.spin_block, self.block_index_spatial, self.inblock_index = splitted

    @property
    def spin_change(self):
        """
        Compute the change in spin induced by an excitation vector element
        referenced by the given tensor element. Assumes that an occupied
        index implies that an electron is taken out of such an object and that
        a virtual index implies that an electron is put into such an object in
        an excitation.
        """
        mapping_spin_change = {
            ("o", "a"): -0.5,  # remove alpha
            ("o", "b"): +0.5,  # remove beta
            ("v", "a"): +0.5,  # add    alpha
            ("v", "b"): -0.5,  # add    beta
        }
        return int(sum(mapping_spin_change[(space[0], spin)]
                       for space, spin in zip(self.subspaces, self.spin_block)))

    def __repr__(self):
        return f"({self.index}  {self.value})"


def find_smallest_matching_elements(predicate, tensor, motrans, n_elements,
                                    degeneracy_tolerance=1e-12):
    """
    Search for the n smallest elements in the passed tensor adhering to the
    passed predicate. If the returned vector contains less elements than
    requested, no more elements matching the criterion can be found. If it
    contains more elements, than the last set of elements beyond the requested
    number have the same value.
    """
    # Search for the n_elements smallest elements in the tensor
    # skipping over elements, which are not fulfilling the passed criteria.
    n_searched_for = max(10, 2 * n_elements + 6)
    while True:
        res = []

        found = tensor.select_n_min(n_searched_for)
        for index, value in found:
            telem = TensorElement(motrans, index, value)
            if predicate(telem):
                res.append(telem)

        if len(res) >= n_elements:
            break  # Everything found

        if len(found) < n_searched_for:
            # We will not be able to find more because already
            # less found than requested
            break
        else:
            n_searched_for *= 2  # Increase for the next round

    if len(res) == 0:
        return []

    # Sort elements in res ascendingly and by spatial indices, i.e. such
    # that blocks differing only in spin will be sorted adjacent
    def telem_nospin(telem):
        return (telem.value, telem.block_index_spatial, telem.inblock_index)

    res = sorted(res, key=telem_nospin)

    # Normalise the tensor values
    istart = 0
    for i in range(len(res)):
        if abs(res[istart].value - res[i].value) > degeneracy_tolerance:
            # Set to the average value
            avg = np.average([r.value for r in res[istart:i]])
            for j in range(istart, i):
                res[j].value = avg
            istart = i
    avg = np.average([r.value for r in res[istart:]])
    for j in range(istart, len(res)):
        res[j].value = avg

    if len(res) > n_elements:
        # Delete the extra elements, excluding the ones of identical value
        return [
            telem for telem in res if telem.value <= res[n_elements - 1].value
        ]
    else:
        return res


def guesses_from_diagonal_singles(matrix, n_guesses, spin_change=0,
                                  spin_block_symmetrisation="none",
                                  degeneracy_tolerance=1e-14,
                                  max_diagonal_value=1000):
    motrans = MoIndexTranslation(matrix.mospaces, matrix.axis_spaces["ph"])
    if n_guesses == 0:
        return []

    # Create a result vector of zero vectors with appropriate symmetry setup
    ret = [guess_zero(matrix, spin_change=spin_change,
                      spin_block_symmetrisation=spin_block_symmetrisation)
           for _ in range(n_guesses)]

    # Search of the smallest elements
    # This predicate checks an index is an allowed element for the singles
    # part of the guess vectors and has the requested spin-change.
    # Also it filters out too large diagonal entries (which are essentially
    # hopeless to give useful excitations)
    def pred_singles(telem):
        return (ret[0].ph.is_allowed(telem.index)
                and telem.spin_change == spin_change
                and abs(telem.value) <= max_diagonal_value)

    elements = find_smallest_matching_elements(
        pred_singles, matrix.diagonal().ph, motrans, n_guesses,
        degeneracy_tolerance=degeneracy_tolerance
    )
    if len(elements) == 0:
        return []

    # By construction of find_smallest_elements the returned elements
    # are already sorted such that adjacent vectors of equal value
    # only differ in spin indices (or they consider excitations from
    # degenerate orbitals). We want to form spin-adapted linear combinations
    # for the case of an unrestricted reference in the following
    # and therefore to exclude the spatial degeneracies we sort explicitly
    # only by value, and spatial indices (and not by spin).

    def telem_nospin(telem):
        return (telem.value, telem.block_index_spatial, telem.inblock_index)

    ivec = 0
    for value, group in groupby(elements, key=telem_nospin):
        if ivec >= len(ret):
            break

        group = list(group)
        if len(group) == 1:  # Just add the single vector
            ret[ivec].ph[group[0].index] = 1.0
            ivec += 1
        elif len(group) == 2:
            # Since these two are grouped together, their
            # spatial parts must be identical.

            # Add the positive linear combination ...
            ret[ivec].ph[group[0].index] = 1
            ret[ivec].ph[group[1].index] = 1
            ivec += 1

            # ... and the negative linear combination
            if ivec < n_guesses:
                ret[ivec].ph[group[0].index] = +1
                ret[ivec].ph[group[1].index] = -1
                ivec += 1
        else:
            raise AssertionError("group size > 3 should not occur "
                                 "when setting up single guesse.")
    assert ivec <= n_guesses

    # Resize in case less guesses found than requested
    # and normalise vectors
    return [evaluate(v / np.sqrt(v @ v)) for v in ret[:ivec]]


def guesses_from_diagonal_doubles(matrix, n_guesses, spin_change=0,
                                  spin_block_symmetrisation="none",
                                  degeneracy_tolerance=1e-14,
                                  max_diagonal_value=1000):
    if n_guesses == 0:
        return []

    # Create a result vector of zero vectors with appropriate symmetry setup
    ret = [guess_zero(matrix, spin_change=spin_change,
                      spin_block_symmetrisation=spin_block_symmetrisation)
           for _ in range(n_guesses)]

    # Build delta-Fock matrices
    spaces_d = matrix.axis_spaces["pphh"]
    df02 = matrix.ground_state.df(spaces_d[0] + spaces_d[2])
    df13 = matrix.ground_state.df(spaces_d[1] + spaces_d[3])

    guesses_d = [gv.pphh for gv in ret]  # Extract doubles parts
    spin_change_twice = int(spin_change * 2)
    assert spin_change_twice / 2 == spin_change
    n_found = libadcc.fill_pp_doubles_guesses(
        guesses_d, matrix.mospaces, df02, df13,
        spin_change_twice, degeneracy_tolerance
    )
    # Resize in case less guesses found than requested
    ret = ret[:n_found]

    # Filter out elements above the noted diagonal value
    diagonal_elements = [ret_d.pphh.dot(matrix.diagonal().pphh * ret_d.pphh)
                         for ret_d in ret]
    return [ret[i] for (i, elem) in enumerate(diagonal_elements)
            if elem <= max_diagonal_value]

# TODO Generic algorithm for forming orthogonal spin components of arbitrary
#      size. Could be useful for the doubles guesses later on.
# if ivec + len(group) <= n_guesses:
#     # Just add all of them with element 1.0
#     for gvec in group:
#         ret[ivec].ph[gvec.index] = 1.0
#         ivec += 1
# else:
#     # Generate orthogonal coefficients for the linear combination
#     # of guess vectors: Make sure every guess vector has a
#     # non-zero component
#     coeff = np.eye(len(group))
#     coeff[:, 0] = 1
#     coeff, _ = np.linalg.qr(coeff)  # orthogonalise

#     for ifrom, ito in enumerate(range(ivec, n_guesses)):
#         for j in range(len(group)):
#             ret[ito].ph[group[j].index] = -coeff[j, ifrom]
#         ivec += 1
